///////////////////////////////////////////////////////////////////////////////
//
// Filename: AncestorMarriageNode.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Representation of a marriage in an Ancestor Diagram.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.ancestor;

import java.awt.*;

public class AncestorMarriageNode
    extends AncestorDiagramNode
{
    private AncestorDiagramNode _husband;
    private AncestorDiagramNode _wife;
    private AncestorPersonNode _child;
    private int _botRowHeight;
    private int _botRowWidth;
    int _husbw;
    
    
    public AncestorMarriageNode( AncestorDiagramNode husband, AncestorDiagramNode wife, AncestorPersonNode child )
    {
        _husband = husband;
        _wife = wife;
        _child = child;
        
        if( (wife == null) && (husband == null) )
            throw new RuntimeException( "Can't have marriage with two nulls" );
        
        _husbw = ( husband != null ? husband.getWidth() : 0 );        
        
        if( husband == null )
        {
            _botRowWidth  = _bounds.width  = wife.getWidth() + 10 + child.getWidth()/2;
            _botRowHeight = _bounds.height = wife.getHeight();
        }
        else if( wife == null )
        {
            _botRowWidth  = _bounds.width  = husband.getWidth() + 10 + child.getWidth()/2;
            _botRowHeight = _bounds.height = husband.getHeight();
        }
        else
        {
            _botRowWidth  = _bounds.width  = husband.getWidth() + wife.getWidth() + 20;
            _botRowHeight = _bounds.height = Math.max( husband.getHeight(), wife.getHeight() );
        }

        _bounds.width = Math.max( _bounds.width, child.getWidth() );        
        _bounds.height = Math.max( _bounds.height, _botRowHeight + child.getHeight() + 20 );
    }

    public int getCentre()
    {
        return _bounds.x + (_bounds.width - _botRowWidth)/2 + _husbw + 8;
    }

    public int getTopRowWidth()
    {
        return (_bounds.width - _botRowWidth)/2 + _botRowWidth;
    }
    
    public void setLocation( int x, int y )
    {
        _bounds.x = x;
        _bounds.y = y;
        
        int trx = (_bounds.width - _botRowWidth)/2;
        
        if( _husband != null )
            _husband.setLocation( x + trx, y + _botRowHeight - _husband.getHeight() );
             
        if( _wife != null )
            _wife.setLocation( x + trx + _husbw + 20, y + _botRowHeight - _wife.getHeight() );
            
        _child.setLocation( this.getCentre() - _child.getCentre() + _child.getX(), _bounds.y + _botRowHeight + 20 );
    }
    
    public void paint(Graphics g, Point origin, double fac, Font nameFont, Font dateFont)
    {
        int bry2 = _bounds.y + _botRowHeight;
        
        if( _husband != null )
            _husband.paint( g, origin, fac, nameFont, dateFont );
                    
        if( _wife != null )
            _wife.paint( g, origin, fac, nameFont, dateFont );
        
        if( (_husband != null) && (_wife != null) )
        {
            // horizontal line between parent centres
            g.drawLine( origin.x + (int)(_husband.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac),
                        origin.x + (int)(_wife.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac) );
        }
        else if( _husband != null )
        {
            // horizontal line between husband centre and child centre
            g.drawLine( origin.x + (int)(_husband.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac),
                        origin.x + (int)(_child.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac) );
        }
        else if( _wife != null )
        {
            // horizontal line between child centre and wife centre
            g.drawLine( origin.x + (int)(_child.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac),
                        origin.x + (int)(_wife.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac) );
        }
        
        // vertical line from father to horizontal line
        if( _husband != null )
        {
            g.drawLine( origin.x + (int)(_husband.getCentre()*fac), origin.y + (int)(bry2*fac),
                        origin.x + (int)(_husband.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac) );
        }
                    
        // vertical line from mother to horizontal line
        if( _wife != null )
        {
            g.drawLine( origin.x + (int)(_wife.getCentre()*fac), origin.y + (int)(bry2*fac),
                        origin.x + (int)(_wife.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac) );
        }
        
        // vertical line from child centre to line
        g.drawLine( origin.x + (int)(_child.getCentre()*fac), origin.y + (int)((bry2 + 10)*fac),
                    origin.x + (int)(_child.getCentre()*fac), origin.y + (int)((bry2 + 20)*fac) );
                    
        _child.paint( g, origin, fac, nameFont, dateFont );

        // DEBUG - draw bounding box in green
        //g.setColor(java.awt.Color.green);
        //g.drawRect(_bounds.x, _bounds.y, _bounds.width, _bounds.height);
        //g.setColor(java.awt.Color.black);
    }
}