///////////////////////////////////////////////////////////////////////////////
//
// Filename: FTDMarriageNode.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Representation of a marriage in a Family Tree Diagram.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.diagram;

import java.awt.Graphics;
import java.awt.Font;
import java.util.Vector;

import java.awt.Point;


public class FTDMarriageNode
    extends FTDiagramNode
{
    private FTDPersonNode _husband;
    private FTDPersonNode _wife;
    private Vector<FTDiagramNode> _children = new Vector<FTDiagramNode>();
    private int _topRowHeight;
    private int _topRowWidth;
    private int _husbw;
    
    public FTDMarriageNode( FTDPersonNode husband, FTDPersonNode wife )
    {
        _husband = husband;
        _wife = wife;

        if( (wife == null) && (husband == null) )
            throw new RuntimeException( "Can't have marriage with two nulls" );
            
        _husbw = ( husband != null ? husband.getWidth() : 0 );
        
        if( husband == null )
        {
            _topRowWidth  = _bounds.width  = wife.getWidth() + 10;
            _topRowHeight = _bounds.height = wife.getHeight();
        }
        else if( wife == null )
        {
            _topRowWidth  = _bounds.width  = husband.getWidth() + 10;
            _topRowHeight = _bounds.height = husband.getHeight();
        }
        else
        {
            _topRowWidth  = _bounds.width  = husband.getWidth() + wife.getWidth() + 20;
            _topRowHeight = _bounds.height = Math.max( husband.getHeight(), wife.getHeight() );
        }
    }
    
    public void addChild( FTDiagramNode child )
    {
        if( _children.isEmpty() )
        {
            _bounds.height = _topRowHeight + child.getHeight();
            _bounds.width = Math.max( _bounds.width, child.getWidth() );
        }
        else
        {
            int childrenWidth = 0;
            int nogrchldWidth = 0;
            
            for( FTDiagramNode node : _children )
            {
                if( node.hasChildren() )
                {
                    nogrchldWidth = childrenWidth + node.getTopRowWidth() + 20;
                    childrenWidth += node.getWidth() + 20;
                }
                else
                {
                    childrenWidth += node.getWidth() + 20;
                    nogrchldWidth += node.getWidth() + 20;
                }
            }
            
            if( child.hasChildren() )
                _bounds.width = Math.max( _bounds.width, childrenWidth + child.getWidth() );
            else
                _bounds.width = Math.max( _bounds.width, nogrchldWidth + child.getWidth() );
                
            _bounds.height = Math.max( _bounds.height, _topRowHeight + child.getHeight() + 60 );
        }
        
        _children.addElement( child );
    }

    public int getCentre()
    {
        return _bounds.x + (_bounds.width - _topRowWidth)/2 + _husbw + 8;
    }

    public int getTopRowWidth()
    {
        return (_bounds.width - _topRowWidth)/2 + _topRowWidth;
    }

    public boolean hasChildren() { return !_children.isEmpty(); }

    public void setLocation( int x, int y )
    {
        // Correction for trees that narrow, then widen
        if( x < 0 )
            FTDiagram.minx = Math.min( x, FTDiagram.minx );
        
        _bounds.x = x;
        _bounds.y = y;
        
        int trx = (_bounds.width - _topRowWidth)/2;

        if( _husband != null )
            _husband.setLocation( x + trx, y );
             
        if( _wife != null )
            _wife.setLocation( x + trx + _husbw + 20, y );

        if( _children.size() > 1 )
        {
            int childrenWidth = 0;
            int nogrchldWidth = 0;
            
            for( FTDiagramNode child : _children )
            {
                if( child.hasChildren() )
                {
                    child.setLocation( _bounds.x + childrenWidth, _bounds.y + _topRowHeight + 20 );
                    nogrchldWidth = childrenWidth + child.getTopRowWidth() + 20;
                    childrenWidth += child.getWidth() + 20;
                }
                else
                {
                    child.setLocation( _bounds.x + nogrchldWidth, _bounds.y + _topRowHeight + 20 );                
                    childrenWidth += child.getWidth() + 20;
                    nogrchldWidth += child.getWidth() + 20;
                }
            }
        }
        else if( !_children.isEmpty() )
        {
            FTDiagramNode child = (FTDiagramNode) _children.firstElement();            
            child.setLocation( this.getCentre() - child.getCentre() + child.getX(), _bounds.y + _topRowHeight + 20 );
        }
    }
    
    public void paint( Graphics g, Point origin, double fac, Font nameFont, Font dateFont )
    {
        int trx = (_bounds.width - _topRowWidth)/2;
        int trm = trx + _husbw + 8;
        int try2 = _bounds.y + _topRowHeight;
        
        if( _husband != null )
            _husband.paint( g, origin, fac, nameFont, dateFont );

        if( (_husband != null) && (_wife != null) )
            g.drawString( "=", origin.x + (int)((_bounds.x + trx + _husband.getWidth() + 8)*fac), origin.y + (int)((_bounds.y + 16)*fac) );
            
        if( _wife != null )
            _wife.paint( g, origin, fac, nameFont, dateFont );
        
        if( !_children.isEmpty() )
        {
            int fcx = ((FTDiagramNode) _children.firstElement()).getCentre();
            int lcx = ((FTDiagramNode) _children.lastElement()).getCentre();
            
            g.drawLine( origin.x + (int)(fcx*fac), origin.y + (int)((try2 + 10)*fac),
                        origin.x + (int)(lcx*fac), origin.y + (int)((try2 + 10)*fac) );
            g.drawLine( origin.x + (int)((_bounds.x + trm)*fac), origin.y + (int)((_bounds.y + 24)*fac),
                        origin.x + (int)((_bounds.x + trm)*fac), origin.y + (int)((try2 + 10)*fac) );

            for( FTDiagramNode node : _children )
            {
                g.drawLine( origin.x + (int)(node.getCentre()*fac), origin.y + (int)((try2 + 10)*fac),
                            origin.x + (int)(node.getCentre()*fac), origin.y + (int)((try2 + 20)*fac) );
                node.paint( g, origin, fac, nameFont, dateFont );
            }
            
            // DEBUG - draw bounding box in green
            //g.setColor(java.awt.Color.green);
            //g.drawRect(_bounds.x, _bounds.y, _bounds.width, _bounds.height);
            //g.setColor(java.awt.Color.black);
        }
    }
}