///////////////////////////////////////////////////////////////////////////////
//
// Filename: GedcomLoader.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Gedcom file loader.
//
// (c) Copyright R.H.Thornburrow, 2001-4
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.gedcom;

import com.tsac.familytree.gui.*;
import com.tsac.familytree.*;
import com.tsac.familytree.util.*;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Calendar;


public class GedcomLoader
    implements FamilyTreeLoader
{
    public static final int NONE  = 0x00;
    public static final int HEAD  = 0x01;
    public static final int HEAD_SOUR = 0x0101;
    public static final int HEAD_DEST = 0x0201;
    public static final int HEAD_DATE = 0x0301;
    public static final int HEAD_FILE = 0x0401;
    public static final int HEAD_GEDC = 0x0501;
    public static final int HEAD_CHAR = 0x0601;
    public static final int HEAD_SUBM = 0x0701;
    public static final int HEAD_SUBN = 0x0801;
    public static final int SUBM  = 0x02;
    public static final int SOUR  = 0x03;
    public static final int REPO  = 0x04;
    public static final int SUBN  = 0x05;
    public static final int INDI  = 0x06;
    public static final int INDI_NAME  = 0x0106;
    public static final int INDI_NAME_GIVN = 0x010106;
    public static final int INDI_NAME_SURN = 0x020106;
    public static final int INDI_AFN   = 0x0206;
    public static final int INDI_SEX   = 0x0306;
    public static final int INDI_SOUR  = 0x0406;
    public static final int INDI_BIRT  = 0x0506;
    public static final int INDI_BIRT_DATE = 0x010506;
    public static final int INDI_BIRT_PLAC = 0x020506;
    public static final int INDI_BIRT_SOUR = 0x030506;
    public static final int INDI_CHR   = 0x0606;
    public static final int INDI_CHR_DATE  = 0x010606;
    public static final int INDI_CHR_PLAC  = 0x020606;
    public static final int INDI_CHR_SOUR  = 0x030606;
    public static final int INDI_DEAT  = 0x0706;
    public static final int INDI_DEAT_DATE = 0x010706;
    public static final int INDI_DEAT_PLAC = 0x020706;
    public static final int INDI_DEAT_SOUR = 0x030706;
    public static final int INDI_DEAT_CAUS = 0x040706;
    public static final int INDI_BURI  = 0x0806;
    public static final int INDI_BURI_DATE = 0x010806;
    public static final int INDI_BURI_PLAC = 0x020806;
    public static final int INDI_BURI_SOUR = 0x030806;
    public static final int INDI_BAPM  = 0x0906;       // Added in 1.1.3 - Baptism
    public static final int INDI_BAPM_DATE = 0x010906; // Added in 1.1.3
    public static final int INDI_BAPM_PLAC = 0x020906; // Added in 1.1.3
    public static final int INDI_BAPM_SOUR = 0x030906; // Added in 1.1.3
    public static final int INDI_EVEN  = 0x0a06;       // Added in 1.1.3 - Misc. Event
    public static final int INDI_EVEN_TYPE = 0x010a06; // Added in 1.1.3
    public static final int INDI_EVEN_DATE = 0x020a06; // Added in 1.1.3
    public static final int INDI_EVEN_PLAC = 0x030a06; // Added in 1.1.3
    public static final int INDI_EVEN_SOUR = 0x040a06; // Added in 1.1.3
    public static final int INDI_SOUR2 = 0x0b06;
    public static final int INDI_FAMS  = 0x0c06;       // XRef to a family spouse record
    public static final int INDI_FAMC  = 0x0d06;       // XRef to a family child record
    public static final int INDI_OCCU  = 0x0e06;       // Occupation
    public static final int INDI_OCCU_DATE = 0x010e06; // Added in 1.1.3
    public static final int INDI_OCCU_PLAC = 0x020e06; // Added in 1.1.3
    public static final int INDI_OCCU_SOUR = 0x030e06; // Added in 1.1.3
    public static final int INDI_NOTE  = 0x0f06;       // XRef to a Note
    public static final int INDI_CHAN  = 0x1006;       // Added in 1.1.3 - Change
    public static final int INDI_CHAN_DATE = 0x011006; // Added in 1.1.3
    public static final int INDI_ADDR  = 0x1106;       // Added in 1.1.3 - Address of Individual
    public static final int INDI_TITL  = 0x1206;       // Added in 1.1.3 - Title
    public static final int INDI_TITL_DATE = 0x011206; // Added in 1.1.3 - Date Title given
    public static final int INDI_IES_DATE  = 0x01ff06; // Individual Event Structure: Date
    public static final int INDI_IES_PLAC  = 0x02ff06; // Individual Event Structure: Place
    public static final int INDI_IES_SOUR  = 0x03ff06; // Individual Event Structure: Source
    public static final int INDI_RESI  = 0x1306;       // Added in 1.1.3-rc4 - Title
    public static final int INDI_MARR  = 0x1406;       // Added in 1.1.3-rc4 - Title
    public static final int FAM   = 0x07;
    public static final int FAM_HUSB   = 0x0107;        // Husband
    public static final int FAM_WIFE   = 0x0207;        // Wife
    public static final int FAM_CHIL   = 0x0307;        // Child
    public static final int FAM_MARR   = 0x0407;        // FES: Marriage structure
    public static final int FAM_MARR_DATE  = 0x010407;
    public static final int FAM_MARR_PLAC  = 0x020407;
    public static final int FAM_MARR_SOUR  = 0x030407;
    public static final int FAM_MARL   = 0x0507;       // Added in 1.1.3 - Marriage Licence
    public static final int FAM_DIV    = 0x0607;       // Divorce
    public static final int FAM_ANUL   = 0x0707;       // Added in 1.1.3 - Anulment
    public static final int FAM_NOTE   = 0x0807;
    public static final int FAM_CHAN   = 0x0907;       // Added in 1.1.3 - Change
    public static final int FAM_CHAN_DATE = 0x010907;  // Added in 1.1.3
    public static final int FAM_FES_DATE  = 0x01ff07;  // Added in 1.1.3
    public static final int FAM_FES_PLAC  = 0x02ff07;  // Added in 1.1.3
    public static final int FAM_FES_SOUR  = 0x03ff07;  // Added in 1.1.3
    public static final int NOTE  = 0x08;              // Added in 1.1.3 - Note
    public static final int TRLR  = 0xfd;
    public static final int USER  = 0xfe;              // Added in 1.1.3 - User-defined tag
    public static final int UNKNOWN = 0xff;            // Added in 1.1.3-rc6 - Unknown tag

    // static GEDCOM tag data
    @SuppressWarnings("unchecked")
    private static Hashtable<String,Integer> _cmds[][] = new Hashtable[0xa][6];
    private static Hashtable<String,Integer> _cmds0 = new Hashtable<String,Integer>();

    // Loader storage
    private Vector<Person> _people = new Vector<Person>();
    private Marriage[] _marriages = null;
    
    // Loader internals
    private Hashtable<String,Integer> _usertags = new Hashtable<String,Integer>();
    private Hashtable<String,Integer> _unknowntags = new Hashtable<String,Integer>();
    private int _mode = NONE;
    private String _param;

    // Individual properties
    private String   _surname;
    private String   _forenames;
    private String   _key;
    private String   _sex;
    private DateBase _birtdate;
    private DateBase _chrdate;
    private DateBase _deatdate;
    private DateBase _buridate;
    private String   _birtplac;
    private String   _chrplac;
    private String   _deatplac;
    private String   _buriplac;

    // Family properties
    private String   _husbkey = null;
    private String   _wifekey = null;
    private Vector<String>   _children = new Vector<String>();
    private DateBase _marrdate = null;
    private String   _marrplace = null;
    
    // Temporary cache of family data
    private Vector<GedcomMarriage> _gedmarrs = new Vector<GedcomMarriage>();
    
    private boolean _loadedOk;
    
    private int _line;
    private LogBuffer _log;
    
    
    /**
     * Private class to hold family data until all the individuals have all
     * been read in - then toMarriage() is used to convert all this data into
     * proper Marriage objects with all the individuals resolved to Person
     * objects.
     */
    private class GedcomMarriage
    {
        private String _husbkey;
        private String _wifekey;
        private Vector<String> _children;
        private DateBase _marrdate;
        private String _marrplace;
        private int _line;
        
        public GedcomMarriage( String husbkey, String wifekey, Vector<String> children,
                               DateBase marrdate, String marrplace, int line )
        {
            _husbkey   = husbkey;
            _wifekey   = wifekey;
            _children  = children;
            _marrdate  = marrdate;
            _marrplace = marrplace;
            _line = line;
        }
        
        Marriage toMarriage()
            throws ParseException
        {
            Person husband = null, wife = null;
            
            if( _husbkey != null )
            {
                if( (husband = Person.getPerson( _husbkey )) == null )
                    _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Invalid husband key '" + _husbkey + "'");
            }
            
            if( _wifekey != null )
            {
                if( (wife = Person.getPerson( _wifekey )) == null )
                    _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Invalid wife key '" + _wifekey + "'");
            }
            
            if( (husband == null) && (wife == null) )
                throw new ParseException( _line + ": Marriage with missing or invalid husband and wife" );
            
            Marriage marriage = new Marriage( husband, wife, _marrdate );
            if( _marrplace != null ) marriage.setPlace( _marrplace );

            if( (_children != null) && !_children.isEmpty() )
            {
                for( String childkey : _children ) {
                    Person child = Person.getPerson( childkey );
                    
                    if( child != null )
                    {
                        if( wife != null )
                        {
                            child.mother = wife;
                            wife.addChild( child );
                        }
                        if( husband != null )
                        {
                            child.father = husband;
                            husband.addChild( child );
                        }
                    }
                    else
                        _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Invalid child key '" + childkey + "'");
                }
            }
            
            return marriage;
        }
    }
    
    
    public GedcomLoader(LogBuffer log)
    {
        _log = log;
        
        _cmds0.put( "HEAD", new Integer( HEAD ) );
        _cmds0.put( "SUBM", new Integer( SUBM ) );
        _cmds0.put( "SOUR", new Integer( SOUR ) );
        _cmds0.put( "REPO", new Integer( REPO ) );
        _cmds0.put( "SUBN", new Integer( SUBN ) );
        _cmds0.put( "INDI", new Integer( INDI ) );
        _cmds0.put( "FAM",  new Integer( FAM  ) );
        _cmds0.put( "NOTE", new Integer( NOTE ) );  // Added in 1.1.3
        _cmds0.put( "TRLR", new Integer( TRLR ) );
        
        for( int n = 0; n < _cmds.length; n++ ) {
            for( int m = 0; m < _cmds[n].length; m++ )
                _cmds[n][m] = new Hashtable<String,Integer>();
        }
        
        _cmds[HEAD][1].put( "SOUR", new Integer( HEAD_SOUR ) ); // Source
        _cmds[HEAD][1].put( "DEST", new Integer( HEAD_DEST ) ); // Destination
        _cmds[HEAD][1].put( "DATE", new Integer( HEAD_DATE ) );
        _cmds[HEAD][1].put( "FILE", new Integer( HEAD_FILE ) );
        _cmds[HEAD][1].put( "GEDC", new Integer( HEAD_GEDC ) ); // Gedcom Information
        _cmds[HEAD][1].put( "CHAR", new Integer( HEAD_CHAR ) ); // Character Set
        _cmds[HEAD][1].put( "SUBM", new Integer( HEAD_SUBM ) ); // Submitter
        _cmds[HEAD][1].put( "SUBN", new Integer( HEAD_SUBN ) ); // Submission
        _cmds[SUBM][1].put( "NAME", new Integer( 0 ) );
        _cmds[SUBM][1].put( "ADDR", new Integer( 0 ) );
        _cmds[SUBM][1].put( "EMAIL", new Integer( 0 ) );        // [1.1.3-rc5]
        _cmds[SOUR][1].put( "AUTH", new Integer( 0 ) );         // Author
        _cmds[SOUR][1].put( "TITL", new Integer( 0 ) );         // Title
        _cmds[SOUR][1].put( "PUBL", new Integer( 0 ) );         // Publication
        _cmds[SOUR][1].put( "REPO", new Integer( 0 ) );         // Repository
        _cmds[SOUR][1].put( "NOTE", new Integer( 0 ) );         // [1.1.3-rc4]
        _cmds[SOUR][1].put( "EMAIL", new Integer( 0 ) );        // [1.1.3-rc5]
        _cmds[REPO][1].put( "NAME", new Integer( 0 ) );
        _cmds[REPO][1].put( "ADDR", new Integer( 0 ) );
        _cmds[REPO][1].put( "EMAIL", new Integer( 0 ) );        // [1.1.3-rc4]
        _cmds[REPO][1].put( "PHON", new Integer( 0 ) );         // [1.1.3-rc4]
        _cmds[SUBN][1].put( "DESC", new Integer( 0 ) );         // Descendants
        _cmds[SUBN][1].put( "ORDI", new Integer( 0 ) );         // (Religious) Ordinance
        _cmds[INDI][1].put( "NAME", new Integer( INDI_NAME ) ); // Person Name Structure
        _cmds[INDI][1].put( "ALIA", new Integer( 0 ) );         // Alias [1.1.3]
        _cmds[INDI][1].put( "TITL", new Integer( 0 ) );         // Title [1.1.3]
        _cmds[INDI][1].put( "AFN",  new Integer( INDI_AFN ) );  // Ancestral File Number
        _cmds[INDI][1].put( "SEX",  new Integer( INDI_SEX ) );  // Sex of Individual
        _cmds[INDI][1].put( "SOUR", new Integer( INDI_SOUR ) ); // Record source
        _cmds[INDI][1].put( "BIRT", new Integer( INDI_BIRT ) ); // IES: Birth Structure
        _cmds[INDI][1].put( "CHR",  new Integer( INDI_CHR ) );  // IES: Chirstening Structure
        _cmds[INDI][1].put( "DEAT", new Integer( INDI_DEAT ) ); // IES: Death Structure
        _cmds[INDI][1].put( "BURI", new Integer( INDI_BURI ) ); // IES: Burial Structure
        _cmds[INDI][1].put( "BAPM", new Integer( INDI_BAPM ) ); // IES: Baptism Structure [1.1.3]
        _cmds[INDI][1].put( "EVEN", new Integer( INDI_EVEN ) ); // IES: Event [1.1.3]
        _cmds[INDI][1].put( "FAMS", new Integer( INDI_FAMS ) ); // Family Spouse (link)
        _cmds[INDI][1].put( "FAMC", new Integer( INDI_FAMC ) ); // Family Child (link)
        _cmds[INDI][1].put( "OCCU", new Integer( INDI_OCCU ) ); // Occupation
        _cmds[INDI][1].put( "NOTE", new Integer( INDI_NOTE ) ); // Textual description
        _cmds[INDI][1].put( "OBJE", new Integer( 0 ) );         // Object (image etc.) [1.1.3]
        _cmds[INDI][1].put( "CHAN", new Integer( INDI_CHAN ) ); // Change [1.1.3]
        _cmds[INDI][1].put( "ADDR", new Integer( INDI_ADDR ) ); // Address of Individual [1.1.3]
        _cmds[INDI][1].put( "TITL", new Integer( INDI_TITL ) ); // Title of Individual [1.1.3]
        _cmds[INDI][1].put( "RESI", new Integer( INDI_RESI ) ); // ? [1.1.3-rc4]
        _cmds[INDI][1].put( "MARR", new Integer( INDI_MARR ) ); // ? [1.1.3-rc4]
        _cmds[INDI][1].put( "SSN", new Integer( 0 ) );          // ? [1.1.3-rc4]
        _cmds[FAM ][1].put( "HUSB", new Integer( FAM_HUSB ) );  // Husband
        _cmds[FAM ][1].put( "WIFE", new Integer( FAM_WIFE ) );  // Wife
        _cmds[FAM ][1].put( "CHIL", new Integer( FAM_CHIL ) );  // Child
        _cmds[FAM ][1].put( "MARR", new Integer( FAM_MARR ) );  // Marriage
        _cmds[FAM ][1].put( "MARL", new Integer( FAM_MARL ) );  // Marriage Licence [1.1.3]
        _cmds[FAM ][1].put( "DIV",  new Integer( FAM_DIV  ) );  // Divorce 
        _cmds[FAM ][1].put( "ANUL", new Integer( FAM_ANUL ) );  // Anulment [1.1.3]
        _cmds[FAM ][1].put( "NOTE", new Integer( FAM_NOTE ) );
        _cmds[FAM ][1].put( "CHAN", new Integer( FAM_CHAN ) );  // Change [1.1.3]
        _cmds[NOTE][1].put( "CONC", new Integer( 0 ) );         // Concatenation [1.1.3]
        _cmds[NOTE][1].put( "CONT", new Integer( 0 ) );         // Continued [1.1.3]
        
        _cmds[HEAD][2].put( "VERS", new Integer( 0 ) );         // Version (of product)
        _cmds[HEAD][2].put( "NAME", new Integer( 0 ) );
        _cmds[HEAD][2].put( "CORP", new Integer( 0 ) );         // Corporate
        _cmds[HEAD][2].put( "ADDR", new Integer( 0 ) );         // [1.1.2]
        _cmds[HEAD][2].put( "DATA", new Integer( 0 ) );         // Data
        _cmds[HEAD][2].put( "TIME", new Integer( 0 ) );
        _cmds[HEAD][2].put( "FORM", new Integer( 0 ) );         // Format
        _cmds[SUBM][2].put( "CONT", new Integer( 0 ) );         // Continued
        _cmds[SOUR][2].put( "CALN", new Integer( 0 ) );         // Call Number
        _cmds[SOUR][2].put( "CONC", new Integer( 0 ) );         // Concatenation [1.1.3-rc4]
        _cmds[REPO][2].put( "CONT", new Integer( 0 ) );         // Continued
        _cmds[INDI][2].put( "GIVN", new Integer( INDI_NAME_GIVN ) ); // Given Names (forenames)
        _cmds[INDI][2].put( "SURN", new Integer( INDI_NAME_SURN ) ); // Surname
        _cmds[INDI][2].put( "DATE", new Integer( INDI_IES_DATE  ) ); // IES: Date for BIRT., CHR., DEAT., BURI., CHAN., TITL.DATE
        _cmds[INDI][2].put( "PLAC", new Integer( INDI_IES_PLAC  ) ); // IES: Place for BIRT., CHR., DEAT. or BURI.PLAC
        _cmds[INDI][2].put( "SOUR", new Integer( INDI_IES_SOUR  ) ); // IES: Source for BIRT., CHR., DEAT. or BURI.SOUR
        _cmds[INDI][2].put( "CAUS", new Integer( INDI_DEAT_CAUS ) ); // Cause of Death
        _cmds[INDI][2].put( "TYPE", new Integer( INDI_EVEN_TYPE ) ); // Event Type [1.1.3]
        _cmds[INDI][2].put( "PAGE", new Integer( 0 ) );         // Page (in a reference work)
        _cmds[INDI][2].put( "CONT", new Integer( 0 ) );         // Continued
        _cmds[INDI][2].put( "CONC", new Integer( 0 ) );         // Concatenation
        _cmds[INDI][2].put( "NOTE", new Integer( 0 ) );         // ? [1.1.3-rc4]
        _cmds[INDI][2].put( "DATA", new Integer( 0 ) );         // ? [1.1.3-rc4]
        _cmds[FAM ][2].put( "DATE", new Integer( FAM_FES_DATE ) );   // FES: Date for MARR., MARL., DIV., ANUL.
        _cmds[FAM ][2].put( "PLAC", new Integer( FAM_FES_PLAC ) );   // FES: Place for MARR., MARL., DIV., ANUL.
        _cmds[FAM ][2].put( "SOUR", new Integer( FAM_FES_SOUR ) );   // FES: Source for MARR., MARL., DIV., ANUL.
        _cmds[FAM ][2].put( "FOST", new Integer( 0 ) );
        
        _cmds[HEAD][3].put( "ADDR", new Integer( 0 ) );
        _cmds[HEAD][3].put( "DATE", new Integer( 0 ) );
        _cmds[HEAD][3].put( "COPR", new Integer( 0 ) );
        _cmds[HEAD][3].put( "PHON", new Integer( 0 ) );             // Phone Number [1.1.3]
        _cmds[SOUR][3].put( "MEDI", new Integer( 0 ) );
        _cmds[INDI][3].put( "TIME", new Integer( 0 ) );             // [1.1.3]
        _cmds[INDI][3].put( "PAGE", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[INDI][3].put( "NOTE", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[INDI][3].put( "DATA", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[INDI][3].put( "TEXT", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[FAM ][3].put( "TIME", new Integer( 0 ) );             // [1.1.3]
        _cmds[FAM ][3].put( "PAGE", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[FAM ][3].put( "DATA", new Integer( 0 ) );             // [1.1.3-rc4]
        
        _cmds[HEAD][4].put( "CONT", new Integer( 0 ) );             // Continued
        _cmds[HEAD][4].put( "PHON", new Integer( 0 ) );             // Phone Number
        _cmds[INDI][4].put( "TEXT", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[INDI][4].put( "CONC", new Integer( 0 ) );             // [1.1.3-rc4]
        _cmds[FAM ][4].put( "TEXT", new Integer( 0 ) );             // [1.1.3-rc4]

        _cmds[INDI][5].put( "CONC", new Integer( 0 ) );             // [1.1.3-rc4]
    }


    public boolean load( String filename )
    {
        _loadedOk = true;
        _mode = NONE;
        _people.removeAllElements();
        _marriages = null;
        _children.removeAllElements();
        
        try
        {
            FileReader frd = new FileReader( filename );
            BufferedReader istrm = new BufferedReader( frd );
            
            String input, cmdstr;
            Integer cmdint;
            int level = -1;
            int index, cmd;
            int userlevel = -1, oldmode = NONE;
            int unknownlevel = -1;
            _line = 0;
                                                
            while( (input = istrm.readLine()) != null )
            {
                _line++;
                
                // Trim whitespace at either end - added in 1.1.3
                input = input.trim();
                
                if( input.length() == 0 )
                    continue;

                try
                {
                    if( (input.length() < 5) || (input.charAt(1) != ' ') )
                        throw new ParseException( "Line too short" );
                
                    try {
                        level = Integer.parseInt( input.substring( 0, 1 ) );
                    } catch( NumberFormatException e ) {
                        throw new ParseException( "Bad level" );
                    }
                    
                    if( input.charAt(2) == '@' )
                    {
                        // line of form "n @xnn@ xxx[x]"
                        index = input.indexOf( " ", 2 );
                            
                        if( index == -1 )
                            throw new ParseException( "Missing @xnn@ command" );
                            
                        cmdstr = input.substring( index+1 );
                        _param = input.substring( 3, index-1 );
                    }
                    else
                    {
                        // line of form "n xxx[x] xxxx..."
                        index = input.indexOf( " ", 2 );
                            
                        if( index == -1 ) {
                            cmdstr = input.substring( 2 );
                            _param = "";
                        } else {    
                            cmdstr = input.substring( 2, index );
                            _param = input.substring( index+1 );
                        }
                    }
                    
                    if( _param.length() == 0 ) _param = null;
                    
                    // User-defined tags start with '_' -  added in 1.1.3
                    if( (cmdstr.length() > 0) && (cmdstr.charAt(0) == '_') )
                    {
                        // Only record first user-defined tag
                        if( _mode != USER )
                        {
                            oldmode = _mode;
                            userlevel = level;
                            setMode( USER );
                        }
                        // or if we move to a higher level
                        else if( level < userlevel )
                        {
                            userlevel = level;
                        }
                        
                        Integer count = (Integer) _usertags.get( cmdstr );
                        
                        if( count == null )
                        {
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Ignoring user-defined tag '" + cmdstr + "'");
                            _usertags.put( cmdstr, new Integer( 1 ) );
                        }
                        else
                        {
                            _usertags.put( cmdstr, new Integer( count.intValue() + 1 ) );
                        }
                        
                        continue;
                    }

                    if( (_mode == USER) || (_mode == UNKNOWN) )
                    {
                        // Added in 1.1.3 to allow for user-defined tags
                        if( level == 0 )
                        {
                            _mode = oldmode;
                            closeRecord();
                            cmdint = (Integer) _cmds0.get( cmdstr );
                        }
                        else
                        {
                            cmdint = (Integer) _cmds[oldmode & 0x0f][level].get( cmdstr );
                            
                            if( cmdint == null )
                                cmdint = new Integer( 0 );
                        }
                    }
                    else
                    {
                        // Normal operation
                        if( level == 0 )
                        {
                            closeRecord();
                            cmdint = (Integer) _cmds0.get( cmdstr );
                        }
                        else
                        {
                            cmdint = (Integer) _cmds[_mode & 0x0f][level].get( cmdstr );
                        }
                    }
                    
                    if( cmdint == null )
                    {
                        if( cmdstr.charAt(0) == '@' )
                            continue;

                        //throw new ParseException( "Unknown tag '" + cmdstr + "' in mode 0x" + Integer.toHexString(_mode) );
                        // Only record first unknown tag
                        if( _mode != UNKNOWN )
                        {
                            oldmode = _mode;
                            unknownlevel = level;
                            setMode( UNKNOWN );
                        }
                        // or if we move to a higher level
                        else if( level < unknownlevel )
                        {
                            unknownlevel = level;
                        }
                        
                        Integer count = (Integer) _unknowntags.get( cmdstr );
                        
                        if( count == null )
                        {
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Ignoring unknown tag '" + cmdstr + "'");
                            _unknowntags.put( cmdstr, new Integer( 1 ) );
                        }
                        else
                        {
                            _unknowntags.put( cmdstr, new Integer( count.intValue() + 1 ) );
                        }

                        continue;
                    }
                    
                    cmd = cmdint.intValue();
                        
                    switch( _mode )
                    {
                    case NONE:
                        if( cmd == HEAD ) setMode( cmd );
                        else throw new ParseException( "Expecting HEAD" );
                        break;
                    case HEAD:
                        if( (cmd == HEAD_SOUR) || (cmd == HEAD_CHAR) ) setMode( cmd );  // HEAD_CHAR added 1.1.3-rc4
                        else throw new ParseException( "Expecting HEAD.SOUR or HEAD.CHAR" );
                        break;
                    case HEAD_SOUR:
                        switch( cmd ) {
                        case HEAD_DEST: case HEAD_DATE: case HEAD_CHAR: case HEAD_GEDC:  // HEAD_CHAR and HEAD_GEDC added 1.1.3-rc4
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting HEAD.DEST, HEAD.DATE, HEAD.CHAR or HEAD.GEDC" );
                            break;
                        } break;
                    case HEAD_DEST:
                        if( cmd == HEAD_DATE ) setMode( cmd );
                        else throw new ParseException( "Expecting HEAD.DATE" );
                        break;
                    case HEAD_DATE:
                        switch( cmd ) {
                        case HEAD_FILE: case HEAD_CHAR: case HEAD_GEDC:  // GEDC added in 1.1.3
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting HEAD.FILE, HEAD.CHAR or HEAD.GEDC" );
                            break;
                        } break;
                    case HEAD_FILE:
                        switch( cmd ) {
                        case HEAD_GEDC:
                        case HEAD_SUBM: // HEAD_SUBM added 1.1.3-rc4
                        case INDI: case FAM: case SOUR: case NOTE:
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting HEAD.GEDC, INDI, FAM, SOUR or NOTE" );
                        } break;
                    case HEAD_GEDC:
                        switch( cmd ) {
                        case HEAD_CHAR: case SUBM: case SUBN:
                        case INDI: case FAM: case SOUR: case NOTE:
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting HEAD.CHAR, SUBM, SUBN, INDI, FAM, SOUR or NOTE" );
                            break;
                        } break;
                    case HEAD_CHAR:
                        switch( cmd ) {
                        case HEAD_SUBM: case HEAD_GEDC: case HEAD_FILE: case HEAD_SOUR:  // HEAD_SOUR added 1.1.3-rc4
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting HEAD.SUBM, HEAD.GEDC, HEAD.FILE or HEAD.SOUR" );
                        } break;
                    case HEAD_SUBM:
                        switch( cmd ) {
                        case HEAD_SUBN: case HEAD_FILE:
                        case HEAD_GEDC: // HEAD_SUBM added 1.1.3-rc4
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting HEAD.SUBN or HEAD.FILE" );
                        } break;
                    case HEAD_SUBN:
                        switch( cmd ) {
                        case SUBM: case SUBN: case REPO: case SOUR:
                        case INDI: case FAM : case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting SUBM, SUBN, REPO, SOUR, INDI, FAM, NOTE or TRLR" );
                        } break;
                    case SUBM:
                        switch( cmd ) {
                        case SUBN: case REPO: case SOUR:
                        case INDI: case FAM : case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting SUBN, REPO, SOUR, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case SOUR:
                        switch( cmd ) {
                        case SUBM: case SUBN: case REPO: case SOUR: // Added SOUR 1.1.3-rc4
                        case INDI: case FAM : case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting SUBM, SUBN, REPO, SOUR, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case REPO:
                        switch( cmd ) {
                        case SUBM: case SUBN: case SOUR: 
                        case INDI: case FAM : case NOTE: case TRLR:
                        case REPO: // REPO added 1.1.3-rc4
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting SUBM, SUBN, SOUR, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case SUBN:
                        switch( cmd ) {
                        case SUBM: case SOUR: case REPO: 
                        case INDI: case FAM: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting SUBM, SOUR, REPO, INDI, FAM, SOUR, NOTE, TRLR" );
                            break;
                        } break;
                                                
                    case INDI: // {0:n}
                        switch( cmd ) {
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting INDI.(NAME/AFN/SEX/SOUR/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR) or INDI, FAM, NOTE, TRLR" );
                        }
                        break;
                    case INDI_NAME: // {1}
                        switch( cmd ) {
                        case INDI_NAME_GIVN: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:
                             setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.NAME.GIVN, INDI.SEX, .SOUR, .BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMC, .FAMS, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE, TRLR" );
                            break;
                        }
                        break;
                    case INDI_NAME_GIVN: // {0:1} - Follows INDI.NAME
                        if( cmd == INDI_NAME_SURN ) setMode( cmd );
                        else if( level < 1 ) throw new ParseException( "Expecting INDI.NAME.SURN" );
                        break;
                    case INDI_NAME_SURN: // {0:1} - Follows INDI.NAME.GIVN
                    case INDI_AFN: // {0:1} - Follows INDI.NAME.SURN
                    case INDI_SEX: // {1}
                        switch( cmd ) {
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                             setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.AFN, .SEX, .SOUR, .BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMC, .FAMS, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE, TRLR" );
                            break;
                        }
                        break;
                    case INDI_SOUR:
                        switch( cmd ) {
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                             setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.SOUR, .BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMC, .FAMS, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE, TRLR" );
                            break;
                        }
                        break;
                        
                    case INDI_BIRT: // {0:1}
                    case INDI_CHR: // {0:1}
                    case INDI_BURI: // {0:1}
                    case INDI_BAPM: // {0:1}
                    case INDI_OCCU: // {0:n}
                        switch( cmd ) {
                        case INDI_IES_DATE:  case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_BIRT: case INDI_CHR: case INDI_BURI: case INDI_BAPM: case INDI_OCCU:
                            if(cmd == _mode) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC/SOUR)" );
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC/SOUR)");
                            setMode( cmd );
                            break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_DEAT:
                        case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC/SOUR)");
                            setMode( cmd );
                            break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC/SOUR)" );
                            break;
                        }
                        break;
                        
                    case INDI_BIRT_DATE: // {1} Expected after BIRT
                    case INDI_CHR_DATE: // {1} Expected after CHR
                    case INDI_BURI_DATE: // {1} Expected after BURI
                    case INDI_BAPM_DATE: // {1} Expected after BAPM
                    case INDI_OCCU_DATE: // {1} Expected after OCCU
                        switch( cmd ) {
                        case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_BIRT: case INDI_CHR: case INDI_BURI: case INDI_BAPM: case INDI_OCCU:
                            if(cmd == _mode) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(PLAC/SOUR)" );
                            setMode( cmd );
                            break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_DEAT:
                        case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(PLAC/SOUR), "
                                    + "INDI.(NAME/AFN/SEX/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), "
                                    + "INDI, FAM, NOTE or TRLR" );
                            break;
                        }
                        break;
                        
                    case INDI_BIRT_PLAC: // {0:1} - Expected after BIRT.DATE
                    case INDI_CHR_PLAC: // {0:1} - Expected after CHR.DATE
                    case INDI_BURI_PLAC: // {0:1} - Expected after BURI.DATE
                    case INDI_BAPM_PLAC: // {0:1} - Expected after BAPM.DATE
                    case INDI_OCCU_PLAC: // {0:1} - Expected after OCCU.DATE
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_BIRT: case INDI_CHR: case INDI_BURI: case INDI_BAPM: case INDI_OCCU:
                            if(cmd == _mode) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/SOUR)" );
                            setMode( cmd );
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_DEAT:
                        case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/SOUR), "
                                    + "INDI.(NAME/AFN/SEX/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), "
                                    + "INDI, FAM, NOTE or TRLR" );
                            break;
                        }
                        break;
                        
                    case INDI_BIRT_SOUR: // {0:1} - Expected after BIRT.DATE or BIRT.PLAC
                    case INDI_CHR_SOUR: // {0:1} - Expected after CHR.DATE or CHR.PLAC
                    case INDI_BURI_SOUR: // {0:1} - Expected after BURI.DATE or BURI.PLAC
                    case INDI_BAPM_SOUR: // {0:1} - Expected after BAPM.DATE or BAPM.PLAC
                    case INDI_OCCU_SOUR: // {0:1} - Expected after OCCU.DATE or OCCU.PLAC
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC:
                            setIESMode( cmd );
                            break;
                        case INDI_BIRT: case INDI_CHR: case INDI_BURI: case INDI_BAPM: case INDI_OCCU:
                            if(cmd == _mode) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC)" );
                            setMode( cmd );
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_DEAT:
                        case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.(BIRT/CHR/BURI/BAPM/OCCU).(DATE/PLAC), "
                                    + "INDI.(NAME/AFN/SEX/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), "
                                    + "INDI, FAM, NOTE or TRLR" );
                            break;
                        }
                        break;

                    case INDI_DEAT: // {0:1}
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode(cmd);
                            break;
                        case INDI_DEAT_CAUS: setMode( cmd ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing INDI.DEAT.(DATE/PLAC/SOUR/CAUS)");
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.DEAT.(DATE/PLAC/SOUR/CAUS)" );
                            break;
                        } break;
                    case INDI_DEAT_DATE: // {1} Expected after DEAT
                        switch( cmd ) {
                        case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode(cmd);
                            break;
                        case INDI_DEAT_CAUS:
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.DEAT.(PLAC/SOUR/CAUS), INDI.(NAME/AFN/SEX/SOUR/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_DEAT_PLAC: // {0:1} - Expected after DEAT.DATE
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_SOUR:
                            setIESMode(cmd);
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_DEAT_CAUS:
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.DEAT.(DATE/SOUR/CAUS), INDI.(NAME/AFN/SEX/SOUR/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_DEAT_SOUR: // {0:1} - Expected after DEAT.DATE or DEAT.PLAC
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC:
                            setIESMode(cmd);
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_DEAT_CAUS:
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.DEAT.(DATE/PLAC/CAUS), INDI.(NAME/AFN/SEX/SOUR/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_DEAT_CAUS: // {0:1} - Expected after DEAT, DEAT.DATE, DEAT.PLAC or DEAT.SOUR
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode(cmd);
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.DEAT.(DATE/PLAC/SOUR), INDI.(NAME/AFN/SEX/SOUR/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                                                                        
                    case INDI_EVEN: // {0:n}
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_EVEN_TYPE: setMode( cmd ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing INDI.EVEN.TYPE");
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.EVEN.TYPE" );
                            break;
                        } break;
                    case INDI_EVEN_TYPE: // {1} - Expected after EVEN
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.EVEN.DATE, .EVEN.PLAC, .EVEN.SOUR, INDI.BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMS, .FAMC, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_EVEN_DATE: // {0:1} - Expected after EVEN.TYPE
                        switch( cmd ) {
                        case INDI_IES_PLAC: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.EVEN.PLAC, .EVEN.SOUR, INDI.BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMS, .FAMC, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_EVEN_PLAC: // {0:1} - Expected after EVEN.TYPE or EVEN.DATE
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_SOUR:
                            setIESMode( cmd );
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.EVEN.SOUR, INDI.SOUR, .BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMS, .FAMC, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_EVEN_SOUR: // {0:1} - Expected after EVEN.TYPE, EVEN.DATE or EVEN.PLAC
                        switch( cmd ) {
                        case INDI_IES_DATE: case INDI_IES_PLAC:
                            setIESMode( cmd );
                            break;
                        case INDI_SOUR: setMode( INDI_SOUR2 ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.SOUR, .BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMS, .FAMC, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                        
                    case INDI_SOUR2:
                        switch( cmd ) {
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.BIRT, .CHR, .BAPM, .DEAT, .BURI, .EVEN, .FAMS, .FAMC, .OCCU, .NOTE, .CHAN, .ADDR, .TITL, INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                    case INDI_FAMS: // {0:n}
                    case INDI_FAMC: // {0:1?}
                    case INDI_NOTE: // {0:1}
                    case INDI_ADDR: // {0:1}
                    case INDI_RESI: // {0:1}  Added 1.1.3-rc4
                    case INDI_MARR: // {0:1}  Added 1.1.3-rc4
                        switch( cmd ) {
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI.(NAME/AFN/SEX/BIRT/CHR/DEAT/BURI/BAPM/EVEN/FAMS/FAMC/OCCU/NOTE/CHAN/ADDR/TITL/RESI/MARR), INDI, FAM, NOTE or TRLR" );
                            break;
                        } break;
                        
                    case INDI_CHAN: // {0:1}
                    case INDI_TITL: // {0:1}
                        switch( cmd ) {
                        case INDI_IES_DATE: setIESMode( cmd ); break;
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing INDI.(CHAN/TITL).DATE");
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting INDI.(CHAN/TITL).DATE" );
                            break;
                        } break;
                    case INDI_CHAN_DATE: // {1} - Follows INDI.CHAN
                    case INDI_TITL_DATE: // {0:1}
                        switch( cmd ) {
                        case INDI_NAME: case INDI_AFN:  case INDI_SEX:  case INDI_SOUR:
                        case INDI_BIRT: case INDI_CHR:  case INDI_DEAT: case INDI_BURI:
                        case INDI_BAPM: case INDI_EVEN: case INDI_FAMS: case INDI_FAMC:
                        case INDI_OCCU: case INDI_NOTE: case INDI_CHAN: case INDI_ADDR:
                        case INDI_TITL: case INDI_RESI: case INDI_MARR:
                        case INDI: case FAM: case NOTE: case TRLR:                            
                            setMode( cmd ); break;
                        default:
                            if( level < 3 ) throw new ParseException( "Expecting INDI.(CHAN/TITL).DATE.TIME" );
                            break;
                        } break;
                                                
                    case FAM:
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR: // empty FAM added 1.1.3-rc4
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting FAM.(HUSB/WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                        }
                        break;
                    case FAM_HUSB: // Order doesn't matter, closeRecord() detects mising WIFE
                        switch( cmd ) {
                        case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting FAM.(WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                            break;
                        }
                        break;
                    case FAM_WIFE: // Order doesn't matter, closeRecord() detects mising WIFE
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting FAM.(HUSB/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                            break;
                        }
                        break;
                    case FAM_CHIL: // {0:M}, Order doesn't matter, closeRecord() detects mising HUSB and WIFE
                        switch( cmd ) {
                        case FAM_CHIL: case FAM_HUSB: case FAM_WIFE:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting FAM.(HUSB/WIFE/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                            break;
                        }
                        break;
                    case FAM_MARR:
                        switch( cmd ) {
                        case FAM_FES_DATE: setMode( FAM_MARR_DATE ); break;
                        case FAM_FES_PLAC: setMode( FAM_MARR_PLAC ); break;
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing FAM.MARR.(DATE/PLAC)");
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting FAM.MARR.DATE or FAM.MARR.PLAC" );
                        } break;
                    case FAM_MARR_DATE: // {0:1} - Expected after MARR
                        switch( cmd ) {
                        case FAM_FES_PLAC: setMode( FAM_MARR_PLAC ); break;
                        case FAM_FES_SOUR: setMode( FAM_MARR_SOUR ); break;
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting FAM.MARR.(PLAC/SOUR), FAM.(WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                        } break;
                    case FAM_MARR_PLAC: // {0:1} - Expected after MARR or MARR.DATE
                        switch( cmd ) {
                        case FAM_FES_DATE: setMode( FAM_MARR_DATE ); break;
                        case FAM_FES_SOUR: setMode( FAM_MARR_SOUR ); break;
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting FAM.MARR.(DATE/SOUR), FAM.(WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                        }
                        break;
                    case FAM_MARR_SOUR: // {0:1} - Expected after MARR.DATE or MARR.PLAC
                        switch( cmd ) {
                        case FAM_FES_DATE: setMode( FAM_MARR_DATE ); break;
                        case FAM_FES_PLAC: setMode( FAM_MARR_PLAC ); break;
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            // level < 2 added 1.1.3-rc4
                            if(level < 2) throw new ParseException( "Expecting FAM.(WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, NOTE or TRLR" );
                        }
                        break;
                    case FAM_MARL: // {0:1}
                        // Stay in this state until we have parsed .DATE, .PLAC and .SOUR
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting FAM.MARL.(DATE/PLAC/SOUR)" );
                            break;
                        }
                        break;
                    case FAM_DIV: // {0:1}
                        // Stay in this state until we have parsed .DATE, .PLAC and .SOUR
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting FAM.DIV.(DATE/PLAC/SOUR)" );
                            break;
                        }
                        break;
                    case FAM_ANUL: // {0:1}
                        // Stay in this state until we have parsed .DATE, .PLAC and .SOUR
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting FAM.ANUL.(DATE/PLAC/SOUR)" );
                            break;
                        }
                        break;
                    case FAM_NOTE: // {0:1}
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            throw new ParseException( "Expecting FAM.(WIFE/CHIL/MARR/MARL/DIV/ANUL/NOTE/CHAN), INDI, FAM, SOUR, NOTE or TRLR" );
                        }
                        break;
                    case FAM_CHAN: // {0:1}
                        switch( cmd ) {
                        case FAM_FES_DATE: setMode( FAM_CHAN_DATE ); break;
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing FAM.CHAN.DATE");
                            setMode( cmd ); break;
                        default:
                            if( level < 2 ) throw new ParseException( "Expecting FAM.CHAN.DATE" );
                            break;
                        } break;
                    case FAM_CHAN_DATE: // {1} Expected after CHAN
                        switch( cmd ) {
                        case FAM_HUSB: case FAM_WIFE: case FAM_CHIL:
                        case FAM_MARR: case FAM_MARL: case FAM_DIV :
                        case FAM_ANUL: case FAM_NOTE: case FAM_CHAN:
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 3 ) throw new ParseException( "Expecting FAM.CHAN.DATE.TIME" );
                            break;
                        } break;
                        
                    case NOTE:
                        switch( cmd ) {
                        case INDI: case FAM: case SOUR: case NOTE: case TRLR:
                            setMode( cmd ); break;
                        default:
                            if( level < 1 ) throw new ParseException( "Expecting INDI, FAM, SOUR, NOTE or TRLR" );
                        } break;
                        
                    case USER:
                        // This code skips all tags until level goes back to that of the user-defined tag
                        if( level <= userlevel ) setMode( cmd );
                        break;

                    case UNKNOWN:
                        // This code skips all tags until level goes back to that of the unknown tag
                        if( level <= unknownlevel ) setMode( cmd );
                        break;
                        
                    default:
                        throw new ParseException( "Unhandled tag '" + cmdstr + "' in mode 0x" + Integer.toHexString(_mode) );
                    }                                        
                }
                catch( NoSuchElementException e )
                {
                    _log.log( LogBuffer.WARNING, "PARSE ERROR: " + _line + ": Not enough commas (" + input + ")" );
                }
                catch( ParseException e )
                {
                    _log.log( LogBuffer.ERROR, "PARSE ERROR: " + _line + ": " + e.getMessage() + " (" + input + ")" );
                    _loadedOk = false;
                    break;
                }
            } /* while */
            
            if( _loadedOk )
            {
                if( _mode == TRLR )
                {
                    try
                    {
                        // Turn all GedcomMarriage records into proper Marriage objects
                        // - this can throw ParseExceptions if we fail to resolve an
                        // individual to a Person object.
                        _marriages = new Marriage[_gedmarrs.size()];
                        for( int n = 0; n < _gedmarrs.size(); n++ )
                            _marriages[n] = ((GedcomMarriage) _gedmarrs.elementAt(n)).toMarriage();
                        
                        System.out.println( "Parsed ok." );
                    }
                    catch( ParseException e )
                    {
                        _log.log(LogBuffer.ERROR, "PARSE ERROR: " + e.getMessage());
                        _loadedOk = false;
                    }
                }
                else
                {
                    _log.log(LogBuffer.ERROR, "PARSE ERROR: Last tag was 0x"
                        + Integer.toHexString(_mode) + ", should be TRLR" );
                    _loadedOk = false;
                }
            }
            
            frd.close();
        }
        catch( IOException e )
        {
            e.printStackTrace();
        }
        
        return _loadedOk;
    }
    
    
    /**
     * Actions to be performed as each tag is recognised.  Actions mainly
     * consist of parsing and remembering any parameters supplied with the tag.
     */
    private void setMode( int mode )
        throws ParseException
    {
        _mode = mode;
        
        // All of these tags don't take parameters
        switch( mode )
        {
        case HEAD:
        case HEAD_GEDC:
        case INDI_BIRT:
        case INDI_CHR:
        case INDI_BAPM:
        case INDI_DEAT:
        case INDI_BURI:
        case INDI_EVEN:
        case INDI_CHAN:
        case INDI_RESI: // Added 1.1.3-rc4
        case INDI_MARR: // Added 1.1.3-rc4
        case FAM_MARR:
        case FAM_MARL:
        case FAM_DIV:
        case FAM_ANUL:
        case FAM_CHAN:
        case TRLR:
        case USER: // ignore all parameters to user-defined tags
        case UNKNOWN: // ignore all parameters to unknown tags
            return;
        case INDI_OCCU:  // parameter is optional?
            return;
        }
        
        // Everything beyond here needs a parameter
        if( _param == null )
        {
            _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing parameter");
            return;
        }
        
        switch( mode )
        {
        case INDI:  _key = _param; break;
        case INDI_NAME:
            if( _param.indexOf('/') != -1 )
            {
                if( _param.indexOf('/') > 0 ) _forenames = _param.substring( 0, _param.indexOf('/') - 1 );
                _surname = _param.substring( _param.indexOf('/') + 1 );
                
                if( _surname.indexOf('/') != -1 ) _surname = _surname.substring( 0, _surname.indexOf('/') );
                if( _surname.indexOf('<') != -1 ) _surname = _surname.substring( 0, _surname.indexOf('<') - 1 );
            }
            else
                _forenames = _param;
            break;
        
        case INDI_NAME_GIVN:
            _forenames = _param;
            break;
        case INDI_NAME_SURN:
            _surname = _param;
            break;

        case INDI_SEX:
            if( _param.equals( "M" ) || _param.equals( "F" ) )
                _sex = _param;
            else
                throw new ParseException( "Sex must be M or F" );
            break;
            
        case INDI_BIRT_DATE:
            if( !_param.equals("UNKNOWN") )
                _birtdate = gedToDateBase(_param, _line);
            break;
        case INDI_CHR_DATE:
            if( !_param.equals("UNKNOWN") )
                _chrdate = gedToDateBase(_param, _line);
            break;
        case INDI_DEAT_DATE:
            if( !_param.equals("UNKNOWN") )
                _deatdate = gedToDateBase(_param, _line);
            break;
        case INDI_BURI_DATE:
            if( !_param.equals("UNKNOWN") )
                _buridate = gedToDateBase(_param, _line);
            break;

        case INDI_BIRT_PLAC:
            _birtplac = _param;
            break;
        case INDI_CHR_PLAC:
            _chrplac = _param;
            break;
        case INDI_DEAT_PLAC:
            _deatplac = _param;
            break;
        case INDI_BURI_PLAC:
            _buriplac = _param;
            break;
        
        case FAM_HUSB:
            _husbkey = _param.substring( 1, _param.length()-1 );
            break;
        case FAM_WIFE:
            _wifekey = _param.substring( 1, _param.length()-1 );
            break;
        case FAM_CHIL:
            _children.addElement( _param.substring( 1, _param.length()-1 ) );
            break;
        case FAM_MARR_DATE:
            if( !_param.equals("UNKNOWN") )
                _marrdate = gedToDateBase( _param, _line );
            break;
        case FAM_MARR_PLAC:
            _marrplace = _param;
            break;
        }
    }
    
    
    /**
     * Convenience method to figure out IES mode from current mode and command
     */
    private void setIESMode(int cmd)
        throws ParseException
    {
        switch(_mode) {
        case INDI_BIRT:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_BIRT_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_BIRT_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_BIRT_SOUR ); break;
            }
            break;
        case INDI_CHR:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_CHR_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_CHR_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_CHR_SOUR ); break;
            }
            break;
        case INDI_DEAT:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_DEAT_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_DEAT_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_DEAT_SOUR ); break;
            }
            break;
        case INDI_BURI:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_BURI_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_BURI_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_BURI_SOUR ); break;
            }
            break;
        case INDI_BAPM:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_BAPM_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_BAPM_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_BAPM_SOUR ); break;
            }
            break;
        case INDI_OCCU:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_OCCU_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_OCCU_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_OCCU_SOUR ); break;
            }
            break;
        case INDI_EVEN:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_EVEN_DATE ); break;
            case INDI_IES_PLAC: setMode( INDI_EVEN_PLAC ); break;
            case INDI_IES_SOUR: setMode( INDI_EVEN_SOUR ); break;
            }
            break;
        case INDI_CHAN:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_CHAN_DATE ); break;
            }
            break;
        case INDI_TITL:
            switch( cmd ) {
            case INDI_IES_DATE: setMode( INDI_TITL_DATE ); break;
            }
            break;
        }
    }
    
    
    /**
     * Have finished parsing a top-level record - need to create an object to
     * represent it.  INDI becomes a Person (Male or Female if sex is known).
     * FAM becomes a GedcomMarriage, which eventually becomes a Marriage.
     */
    private void closeRecord()
        throws ParseException
    {
        switch( _mode & 0x0f )
        {
        case INDI:
            if( _key == null )
                throw new ParseException( "Missing key for individual" );
                
            if( (_surname == null) && (_forenames == null) )
                _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing name for individual");

            Person person = null;
            
            if( (_sex != null) && _sex.equals( "M" ) )
                person = new Male( _key, _surname, _forenames );
            else if( (_sex != null) && _sex.equals( "F" ) )
                person = new Female( _key, _surname, _forenames );
            else
            {
                _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing sex for individial");
                person = new Person( _key, _surname, _forenames );
            }
                
            if( person != null )
            {
                if( _birtdate != null )
                    person.setBorn( _birtdate );
                if( _birtplac != null )
                    person.setBornPlace( _birtplac );
                    
                if( _chrdate != null )
                    person.setChristened( _chrdate );
                if( _chrplac != null )
                    person.setBornPlace( _chrplac );

                if( _deatdate != null )
                    person.setDied( _deatdate );
                if( _deatplac != null )
                    person.setDiedPlace( _deatplac );
                    
                if( (_buridate != null) && (_deatdate == null) )
                    person.setDied( _buridate );
                if( (_buriplac != null) && (_deatplac == null) )
                    person.setDiedPlace( _buriplac );

                _people.addElement( person );
            }

            _forenames = null; _surname = null;
            _key = null; _sex = null;
            _birtdate = null; _birtplac = null;
            _chrdate  = null; _chrplac  = null;
            _deatdate = null; _deatplac = null;
            _buridate = null; _buriplac = null;
            break;
            
        case FAM:
            if( (_husbkey != null) || (_wifekey != null) )
            {
                GedcomMarriage gedmarr;
                
                if( !_children.isEmpty() )
                {
                    gedmarr = new GedcomMarriage( _husbkey, _wifekey,
                        _children, _marrdate, _marrplace, _line );
                    _children = new Vector<String>();
                }
                else
                    gedmarr = new GedcomMarriage( _husbkey, _wifekey,
                        null, _marrdate, _marrplace, _line );
                
                _gedmarrs.addElement( gedmarr );
            }
            else
            {
                if( _husbkey == null )
                    _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing FAM.HUSB, can't create marriage");
                    
                if( _wifekey == null )
                    _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": Missing FAM.WIFE, can't create marriage");
                
                if( !_children.isEmpty() )
                    _log.log(LogBuffer.WARNING, "WARNING: " + _line + ": FAM.CHIL data will not be stored");
                    
                _children.removeAllElements();
            }
            
            _husbkey = null;
            _wifekey = null;
            _marrdate = null;
            _marrplace = null;
            break;
        }
    }

    
    public Marriage[] getMarriages()
    {
        if( _marriages != null )
            return _marriages;
        
        return new Marriage[0];
    }
    
    public Person[] getPeople()
    {
        Person[] ret = new Person[_people.size()];
        _people.copyInto( ret );
        return ret;
    }
    
    public Person[] getConventionalRoots()  { return null; }
    public Person[] getInvertedRoots()  { return null; }
    public boolean isLoadedOk()  { return _loadedOk; }
    
    
    public static void main( String[] args )
    {
        if( args.length != 1 )
        {
            System.err.println( "Usage: ged2ftml <gedcom_file>" );
            System.exit( 1 );
        }
        
        String infile = args[0];
        String outfile;
        
        if( infile.regionMatches( true, infile.length() - 4, ".ged", 0, 4 ) )
            outfile = infile.substring( 0, infile.length() - 4 ) + ".xml";
        else
            outfile = infile + ".xml";
                    
        try
        {
            System.out.println( "Parsing \"" + infile + "\"..." );
            
            GedcomLoader loader = new GedcomLoader(new LogBuffer());
            loader.load( infile );
            
            Person[] people = loader.getPeople();
            Marriage[] marriages = loader.getMarriages();
            
            if( loader.isLoadedOk() )
            {
                System.out.println( "Writing \"" + outfile + "\"..." );
                
                PrintWriter pout = new PrintWriter( new FileWriter( outfile ) );
                //PrintWriter pout = new PrintWriter( System.out );
                pout.println( "<?xml version=\"1.0\" encoding=\"latin1\" ?>" );
                pout.println( "" );
                pout.println( "<!DOCTYPE ftml SYSTEM \"ftml.dtd\">" );
                pout.println( "" );
                pout.println( "<ftml>" );
                pout.println( " <people>" );
                
                for( int n = 0; n < people.length; n++ )
                    people[n].write( pout );

                pout.println( " </people>" );
                pout.println( " <marriages>" );
                
                for( int n = 0; n < marriages.length; n++ )
                    marriages[n].write( pout );

                pout.println( " </marriages>" );
                pout.println( "</ftml>" );
                pout.flush();
                pout.close();
            }
        }
        catch( Throwable e )
        {
            e.printStackTrace();
        }
        
        // For VisCafe only
        //System.out.println( "Press any key to exit" );
        //try { System.in.read(); } catch( java.io.IOException e ) {}
        //System.exit( 0 );
    }
    
    private DateBase gedToDateBase( String str, int line )
    {
        String upcase = str.toUpperCase();
        
        try {
            if( upcase.startsWith( "BEF " ) || upcase.startsWith( "BEF. " ) )
                return DateUtil.toDateBase(
                        str.substring( str.indexOf(' ') + 1 ), ApproxDate.BEFORE, line );
            else if( upcase.startsWith( "AFT " ) || upcase.startsWith( "AFT. " ) )
                return DateUtil.toDateBase(
                        str.substring( str.indexOf(' ') + 1 ), ApproxDate.AFTER, line );
            else if( upcase.startsWith( "ABT " ) || upcase.startsWith( "ABT. " ) )
                return DateUtil.toDateBase(
                        str.substring( str.indexOf(' ') + 1  ), ApproxDate.ABOUT, line );
            else if( upcase.startsWith( "BET " ) || upcase.startsWith( "BET. " ) ) {
                String dates = str.substring( str.indexOf(' ') + 1  );
                String tostr = null, fromstr = null;
                int dash = dates.indexOf( '-' );
                int andPos = -1;

                if( dash != -1 ) {
                    // Split the date on the "-"
                    tostr   = dates.substring( dash + 1 ).trim();
                    fromstr = dates.substring( 0, dash ).trim();                
                } else {
                    andPos = dates.toUpperCase().indexOf( "AND" );
                    if( andPos != -1 ) {
                        // Split the date on the "and"
                        tostr   = dates.substring( andPos + 3 ).trim();
                        fromstr = dates.substring( 0, andPos ).trim();                
                    }
                }

                if( (dash != -1) || (andPos != -1) ) {
                    // Parse the second date
                    DateBase to  = DateUtil.toDateBase( tostr, ApproxDate.NONE, line );

                    // Of the from "25 - 26 Jun 2004"
                    if( to instanceof AbsoluteDate ) {
                        // First number must have 1 or 2 digits
                        if( fromstr.length() <= 2 ) {
                            try {
                                int fromdate = Integer.parseInt( fromstr );
                                if( (fromdate > 0) && (fromdate <= 31) ) {
                                    Calendar fromcal = (Calendar) ((AbsoluteDate) to).getDate().clone();
                                    fromcal.set( Calendar.DATE, fromdate );
                                    DateBase from = new AbsoluteDate( fromcal );
                                    return new DateRange( from, to );
                                }
                            }
                            catch( NumberFormatException e ) {
                            }
                        }


                        _log.log(LogBuffer.WARNING, "WARNING: Unable to make sense of date range \""
                                + dates + "\"");   
                        return null;
                    }

                    // Of the form "2003 - 2004" or "11/2003 - 12/2003"
                    DateBase from = DateUtil.toDateBase( fromstr, ApproxDate.NONE, line );

                    return new DateRange( from, to );
                }

                _log.log(LogBuffer.WARNING, "WARNING: Missing '-' in date range \"" + dates + "\"");
                return null;
            }

            return DateUtil.toDateBase( str, ApproxDate.NONE, line );
        } catch(DateException e) {
            _log.log(LogBuffer.WARNING, e.getMessage());
        }
        
        return null;
    }
}