///////////////////////////////////////////////////////////////////////////////
//
// Filename: TLDiagram.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Time Line diagram.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.timeline;

import com.tsac.familytree.*;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.event.MouseMotionListener;
import java.util.ResourceBundle;

@SuppressWarnings("serial")
public class TLDiagram
    extends javax.swing.JComponent
{
    static ResourceBundle res = ResourceBundle.getBundle("Timeline");
    private String _title = res.getString("Timeline");
    private Person[] _people;
    private int _minYear;
    private int _maxYear;
    private Rectangle _bounds = new Rectangle();
    private Font _font      = new Font( "Helvetica", Font.PLAIN, 9 );
    private Font _titleFont = new Font( "TimesRoman", Font.ITALIC, 24 );
    private FontMetrics _metrics      = getFontMetrics( _font );
    private FontMetrics _titleMetrics = getFontMetrics( _titleFont );

    private TLSignificantEvent[] _sigEvents = {
        new TLSignificantEvent( 1841, 0, res.getString("Census_1841") ),
        new TLSignificantEvent( 1871, 0, res.getString("Census_1871") ),
        new TLSignificantEvent( 1901, 0, res.getString("Census_1901") ),
        new TLSignificantEvent( 1914, 4, res.getString("World_War_I") ),
        new TLSignificantEvent( 1939, 5, res.getString("World_War_II") )
    };

    private static final int XSCALE = 3;
    private static int YSCALE = 5;
    private static int YBASE = 4;
    private static final int XINC = 10;

    public TLDiagram()
    {
        this.setBackground( Color.white );

        // Add the tooltip thingy
        this.addMouseMotionListener( new MouseMotionListener() {
            public void mouseDragged( java.awt.event.MouseEvent e )  {}
            public void mouseMoved( java.awt.event.MouseEvent e )  {
                TLSignificantEvent found = null;
                if( (e.getX() > _bounds.x) && (e.getX() < (_bounds.x + _bounds.width))
                    && (e.getY() > _bounds.y) && (e.getY() < (_bounds.y + _bounds.height)) )
                {
                    int year = _minYear + (e.getX() - _bounds.x)/XSCALE;

                    for( TLSignificantEvent se : _sigEvents ) {
                        if( se.isInRange( year ) ) {
                            found = se;
                            break;
                        }
                    }
                    // Not found - try one pixel either side
                    if(found == null) {
                        year = _minYear + (e.getX() - _bounds.x - 1)/XSCALE;
                        for( TLSignificantEvent se : _sigEvents ) {
                            if( se.isInRange( year ) ) {
                                found = se;
                                break;
                            }
                        }
                    }
                    if(found == null) {
                        year = _minYear + (e.getX() - _bounds.x + 1)/XSCALE;
                        for( TLSignificantEvent se : _sigEvents ) {
                            if( se.isInRange( year ) ) {
                                found = se;
                                break;
                            }
                        }
                    }
                }

                setToolTipText( ( found != null ) ? found.getName() : null );
            }
        } );

        this.setToolTipText( " " );
    }

    public void setData( Person[] people )
    {
        _people = people;

        YSCALE = _metrics.getHeight(); // + metrics.getDescent();
        YBASE = _metrics.getAscent() - 2;

        int minYear = 9999;
        int maxYear = 0;
        int namewidth = 140;

        for( Person person : people )
        {
            int bornYear = person.getBornYear();
            if( (bornYear != 0) && (bornYear != -1) ) {
                minYear = Math.min( minYear, bornYear );
                maxYear = Math.max( maxYear, bornYear );
            }

            int diedYear = person.getDiedYear();
            if( (diedYear != 0) && (diedYear != -1) ) {
                maxYear = Math.max( maxYear, diedYear );
                minYear = Math.min( minYear, diedYear );
            }

            if( _metrics.stringWidth( person.getName() ) > namewidth )
                namewidth = _metrics.stringWidth( person.getName() );
        }

        _minYear = minYear - (minYear % XINC);
        _maxYear = maxYear - (maxYear % XINC) + XINC;

        _bounds.y = _titleMetrics.getHeight() * 2;
        _bounds.x = namewidth + XSCALE*2;
        _bounds.width = (_maxYear - _minYear) * XSCALE;
        _bounds.height = people.length * YSCALE;

        Dimension size = new Dimension( _bounds.x + _bounds.width + _metrics.stringWidth("XXXX") + XSCALE, _bounds.y + 32 + _bounds.height );
        this.setMaximumSize( size );
        this.setPreferredSize( size );
    }


    public void paint( Graphics g, double fac, Font font )
    {
        // draw background
        //g.setColor( this.getBackground() );
        //Dimension size = this.getSize();
        //g.fillRect( 0, 0, size.width, size.height );

        // We don't scale the y-position because of the title
        Rectangle facbounds = new Rectangle( (int)(_bounds.x * fac), _bounds.y, (int)(_bounds.width * fac), (int)(_bounds.height*fac) );

        // title
        g.setColor( Color.black );
        g.setFont( _titleFont );
        g.drawString( _title, facbounds.width/2 - _titleMetrics.stringWidth(_title)/2, _titleMetrics.getHeight()/2 + _titleMetrics.getAscent() );

        // axes
        g.setColor( Color.black );
        g.drawLine( facbounds.x, facbounds.y, facbounds.x, facbounds.y + facbounds.height ); // y-axis
        g.drawLine( facbounds.x, facbounds.y + facbounds.height, facbounds.x + facbounds.width, facbounds.y + facbounds.height ); // x-axis

        g.setFont( font );

        // scales
        int facy = (int)((_bounds.y + _bounds.height) * fac);
        for( int year = _minYear; year <= _maxYear; year += XINC )
        {
            int facx = (int)((_bounds.x + (year - _minYear) * XSCALE) * fac);
            g.drawLine( facx, facy, facx, facy+2 ); // x-grad

            if( (year % (XINC*XINC)) == 0 )
                g.drawString( Integer.toString( year ), facx, facy + (int)(YSCALE*fac) ); // x-scale
        }

        // significant events
        g.setColor( Color.gray );
        for( TLSignificantEvent se : _sigEvents ) {
            if( (se.getYear() >= _minYear) && (se.getYear() <= _maxYear) ) {
                int facx = (int)((_bounds.x + (se.getYear() - _minYear)*XSCALE) * fac);

                if( se.getDuration() > 0 )
                    g.fillRect( facx, facbounds.y, (int)(se.getDuration()*XSCALE*fac), facbounds.height ); // y-mark
                else
                    g.drawLine( facx, facbounds.y, facx, facy ); // y-mark
            }
        }

        // people and lines
        if( _people != null )
        {
            for( int n = 0; n < _people.length; n++ )
            {
                facy = (int)((_bounds.y + n*YSCALE) * fac);

                g.setColor( Color.black );

                // divider line between roots
                //if( _people[n].isRoot() )
                //    g.drawLine( facbounds.x, facy, facbounds.x + facbounds.width, facy );

                g.drawString( _people[n].getName(), XSCALE, facy + YBASE );

                if( (_people[n].getBornYear() != 0) && (_people[n].getDiedYear() != 0) )
                {
                    int facx = (int)((_bounds.x + (_people[n].getBornYear() - _minYear) * XSCALE) * fac);
                    g.setColor( Color.blue );
                    if( _people[n].getAge() > 0 )
                        g.fillRect( facx, facy, (int)(_people[n].getAge() * XSCALE * fac), (int)((YSCALE-3)*fac) );
                    else
                        g.drawLine( facx, facy, facx, facy + (int)((YSCALE-3)*fac) );
                }
                else if( _people[n].getBornYear() != 0 )
                {
                    int facx = (int)((_bounds.x + (_people[n].getBornYear() - _minYear) * XSCALE) * fac);
                    g.setColor( Color.blue );
                    g.fillRect( facx, facy, (int)((YSCALE-3)*fac), (int)((YSCALE-3)*fac) );
                }
                else if( _people[n].getDiedYear() != 0 )
                {
                    int facx = (int)((_bounds.x + (_people[n].getDiedYear() - _minYear) * XSCALE) * fac);
                    g.setColor( Color.black );
                    g.fillRect( facx, facy, (int)((YSCALE-3)*fac), (int)((YSCALE-3)*fac) );
                }
            }
        }
    }


    public void paint( Graphics g )
    {
        this.paint( g, 1.0, _font );
    }


    public void print( Graphics g, Dimension size )
    {
        double xfac = (double) size.width / (double) (_bounds.x + _bounds.width + _metrics.stringWidth("XXXX") + XSCALE);
        double yfac = (double) size.height / (double) (_bounds.y + _bounds.height + YSCALE*2);
        double fac = Math.min( xfac, yfac );

        Font font = new Font( "Helvetica", Font.PLAIN, (int)(9 * fac) );

        this.paint( g, fac, font );
    }
}