///////////////////////////////////////////////////////////////////////////////
//
// Filename: DateBase.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Base class for Date representations.
//
// (c) Copyright R.H.Thornburrow, 2001-4
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.util;

import java.util.ResourceBundle;


public abstract class DateBase
{
    static ResourceBundle res = ResourceBundle.getBundle("FamilyTree");

    public static int difference( DateBase start, DateBase end )
    {
        // If both absolute, calculate absolute difference
        if( (start instanceof AbsoluteDate) && (end instanceof AbsoluteDate) )
            return DateUtil.difference( ((AbsoluteDate) start).getDate(),
                                        ((AbsoluteDate) end).getDate() );
        
        // If start is Before, After or About, then can't calculate
        if( (start instanceof ApproxDate)
            && (((ApproxDate) start).getBeforeAfterAbout() != ApproxDate.NONE) )
            return -1;

        // If end is Before, After or About, then can't calculate
        if( (end instanceof ApproxDate)
            && (((ApproxDate) end).getBeforeAfterAbout() != ApproxDate.NONE) )
            return -1;

        // If both dates specify a month, calculate reasonably exact difference        
        if( (start.getMonth() != -1) && (end.getMonth() != -1) )
            return ((end.getYear() * 12 + end.getMonth())
                - (start.getYear() * 12 + start.getMonth()))/12;
        
        // Only accurate to a whole year, so difference between years
        return (end.getYear() - start.getYear());
    }
    
    
    public static String differenceString( DateBase start, DateBase end )
    {
        // If both absolute, calculate absolute difference
        if( (start instanceof AbsoluteDate) && (end instanceof AbsoluteDate) )
            return DateUtil.differenceString(
                ((AbsoluteDate) start).getDate(),
                ((AbsoluteDate) end).getDate() );
        
        int startBefAftAbt = ApproxDate.NONE;
        int endBefAftAbt   = ApproxDate.NONE;
        
        // Get start Before, After or About setting
        if( start instanceof ApproxDate )
            startBefAftAbt = ((ApproxDate) start).getBeforeAfterAbout();

        // Get end Before, After or About setting
        if( end instanceof ApproxDate )
            endBefAftAbt = ((ApproxDate) end).getBeforeAfterAbout();

        // Both exact(ish) dates, so calculate difference, at best in months
        if( (startBefAftAbt == ApproxDate.NONE) && (endBefAftAbt == ApproxDate.NONE) )
            return monthDifferenceString( start, end );

        // One or both dates About, so difference is About
        if( ((startBefAftAbt == ApproxDate.ABOUT) && (endBefAftAbt == ApproxDate.NONE))
            || ((startBefAftAbt == ApproxDate.NONE) && (endBefAftAbt == ApproxDate.ABOUT))
            || ((startBefAftAbt == ApproxDate.ABOUT) &&  (endBefAftAbt == ApproxDate.ABOUT)) )
        {
            return res.getString("About") + " " + monthDifferenceString( start, end );
        }

        // Start is Before and/or end is After, so difference is a minimum of
        if( ((startBefAftAbt == ApproxDate.BEFORE) && (endBefAftAbt == ApproxDate.NONE))
            || ((startBefAftAbt == ApproxDate.NONE) && (endBefAftAbt == ApproxDate.AFTER))
            || ((startBefAftAbt == ApproxDate.BEFORE) &&  (endBefAftAbt == ApproxDate.AFTER)) )
        {
            return res.getString("At_least") + " " + monthDifferenceString( start, end );
        }

        // Start is After and/or end is Before, so difference is a maximum of
        if( ((startBefAftAbt == ApproxDate.AFTER) && (endBefAftAbt == ApproxDate.NONE))
            || ((startBefAftAbt == ApproxDate.NONE) && (endBefAftAbt == ApproxDate.BEFORE))
            || ((startBefAftAbt == ApproxDate.AFTER) &&  (endBefAftAbt == ApproxDate.BEFORE)) )
        {
            return res.getString("At_most") + " " + monthDifferenceString( start, end );
        }

        // Anything else doesn't make sense
        return null;
    }
    
    
    private static String monthDifferenceString( DateBase start, DateBase end )
    {
        if( (start.getYear() == -1) || (end.getYear() == -1)
            || (end.getYear() < start.getYear()) )
            return null;
            
        // If both dates specify a month, calculate reasonably exact difference        
        if( (start.getMonth() != -1) && (end.getMonth() != -1) )
        {
            int months = ((end.getYear() * 12 + end.getMonth())
                         - (start.getYear() * 12 + start.getMonth()));
                         
            // Anything less then 2 year is reported in months
            if( months < 24 )
                return months + " " + res.getString("months");
                
            // Else just report the number of whole years
            return Integer.toString( months / 12 );
        }
        
        // Only accurate to a whole year, so difference between years
        return Integer.toString( end.getYear() - start.getYear() );
    }
    
    
    public abstract String toString();
    
    public abstract int getMonth();
    
    public abstract int getYear();
    
    public abstract String getYearString();
}