///////////////////////////////////////////////////////////////////////////////
//
// Filename: GedcomXmlHandler.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  27th February 2010
//
// Purpose:  Gedcom XML SAX Handler for use with Xerces.
//
// (c) Copyright R.H.Thornburrow, 2001-10
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.xml.ged2x;

import com.tsac.familytree.*;
import com.tsac.familytree.util.DateBase;
import com.tsac.familytree.util.DateException;
import com.tsac.familytree.util.LogBuffer;
import com.tsac.familytree.xml.IXmlHandlerDelegate;
//import com.tsac.familytree.util.*;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import java.util.Hashtable;
import java.util.Vector;

public class Ged2XXmlHandler
    extends org.xml.sax.helpers.DefaultHandler
    implements IXmlHandlerDelegate
{
    private Hashtable<String,Integer> _elements = new Hashtable<String,Integer>();
    private Hashtable<String,Person>  _people    = new Hashtable<String,Person>();
    private Vector<Person>   _roots     = new Vector<Person>();
    private Hashtable<String,Marriage> _marriages = new Hashtable<String,Marriage>();
    
    private Hashtable<String,Ged2XPerson> _gedPeople   = new Hashtable<String,Ged2XPerson>();
    
    private Ged2XPerson       _currPerson = null;
    private Ged2XmlEvent        _currEvent  = null;
    private Ged2XRelationship _currRelationship = null;
    
    private String    _tempVal   = null;
    private LogBuffer _log;
    
    public static final int FAMILY_TREE  = 1;
    public static final int PERSON       = 2;
    public static final int SEX          = 3;
    public static final int NAME         = 4;
    public static final int GIVEN_NAMES  = 5;
    public static final int FAMILY_NAME  = 6;
    public static final int EVENT        = 7;
    public static final int DATE         = 8;
    public static final int RELATIONSHIP = 9;
     
    
    private class Ged2XPerson
    {
        public String id;
        public String given_names;
        public String family_name;
        public String sex;
        private Vector<Ged2XmlEvent> events = new Vector<Ged2XmlEvent>();
        private Vector<Ged2XRelationship> relationships = new Vector<Ged2XRelationship>();
    }
        
    private class Ged2XmlEvent
    {
        public String type;
        public String family;
        public String date;
    }

    private class Ged2XRelationship
    {
        public String type;
        public String linkto;
        public String family;
    }
    
    /**
     * Constructor.
     */
    public Ged2XXmlHandler(LogBuffer log)
    {
        _log = log;
        
        _elements.put( "family_tree",   new Integer( FAMILY_TREE ) );
        _elements.put( "person",        new Integer( PERSON ) );
        _elements.put( "sex",           new Integer( SEX ) );
        _elements.put( "name",          new Integer( NAME ) );
        _elements.put( "given_names",   new Integer( GIVEN_NAMES ) );
        _elements.put( "family_name",   new Integer( FAMILY_NAME ) );
        _elements.put( "event",         new Integer( EVENT ) );
        _elements.put( "date",          new Integer( DATE ) );
        _elements.put( "relationship",  new Integer( RELATIONSHIP ) );
    }


    public Vector<Person> getPeople() {
        return new Vector<Person>(_people.values());
    }
    public Vector<Person> getRoots() {
        return _roots;
    }
    public Vector<Marriage> getMarriages() {
        return new Vector<Marriage>(_marriages.values());
    }


    @Override
    public void endDocument()
        throws SAXException
    {
        // Create all the people objects
        for(Ged2XPerson gp : _gedPeople.values()) {
            Person person;

            if( (gp.sex != null) && "M".equalsIgnoreCase( gp.sex ) ) {
                person = new Male( gp.id, gp.family_name, gp.given_names );
            } else if( (gp.sex != null) && "F".equalsIgnoreCase( gp.sex ) ) {
                person = new Female( gp.id, gp.family_name, gp.given_names );
            } else {
                _log.log(LogBuffer.WARNING, "WARNING: Missing sex for individual " + gp.id);
                person = new Person( gp.id, gp.family_name, gp.given_names );
            }
            
            // Set birth/death dates
            for(Ged2XmlEvent ge : gp.events) {
                if((ge.type != null) && (ge.date != null)) {
                    try {
                        if("birth".equalsIgnoreCase(ge.type)) {
                            person.setBorn(ge.date);
                        } else if("death".equalsIgnoreCase(ge.type)) {
                            person.setDied(ge.date);
                        }
                    } catch(DateException e) {
                        _log.log(LogBuffer.WARNING, e.getMessage());
                    }
                }
            }
            
            //System.out.println("Person: " + gp.id);
            _people.put( gp.id, person );            
        }
        
        // Create parent/child relationships
        for(Ged2XPerson gp : _gedPeople.values()) {
            for(Ged2XRelationship gr : gp.relationships) {
                if(gr.linkto != null) {
                    String other = extractId(gr.linkto);
                    if("parentof".equalsIgnoreCase(gr.type)) {
                        Person child = _people.get(other);
                        if(child != null) {
                            if("M".equalsIgnoreCase(gp.sex)) {
                                child.setFather(gp.id);
                            } else if("F".equalsIgnoreCase(gp.sex)) {
                                child.setMother(gp.id);
                            }
                        } else {
                            _log.log(LogBuffer.WARNING, "WARNING: Child " + other + " not found for " + gp.id);
                        }
                    } else if("childofmother".equalsIgnoreCase(gr.type)) {
                        Person child = _people.get(gp.id);
                        Person mother = _people.get(other);
                        if((child != null) && (mother != null)) {
                            child.setMother(other);
                        } else {
                            _log.log(LogBuffer.WARNING, "WARNING: Mother " + other + " not found for child " + gp.id);
                        }
                    } else if("childoffather".equalsIgnoreCase(gr.type)) {
                        Person child = _people.get(gp.id);
                        Person father = _people.get(other);
                        if((child != null) && (father != null)) {
                            child.setFather(other);
                        } else {
                            _log.log(LogBuffer.WARNING, "WARNING: Father " + other + " not found for child " + gp.id);
                        }
                    }
                }
            }
        }

        // Create all the marriages
        for(Ged2XPerson gp : _gedPeople.values()) {
            Person person = _people.get(gp.id);
            for(Ged2XRelationship gr : gp.relationships) {
                if("spouseof".equalsIgnoreCase(gr.type)) {
                    if(gr.family != null) {
                        Marriage marr = _marriages.get(gr.family);
                        if(marr != null) {
                            if(gp.sex != null) {
                                if("M".equalsIgnoreCase(gp.sex)) {
                                    if(marr.husband == null) {
                                        marr.husband = person;
                                    } else if(marr.husband != person) {
                                        _log.log(LogBuffer.WARNING, "SEMANTIC ERROR: Husband does not match existing entry for " + gr.family);
                                    }
                                } else if("F".equalsIgnoreCase(gp.sex)) {
                                    if(marr.wife == null) {
                                        marr.wife = person;
                                    } else if(marr.wife != person) {
                                        _log.log(LogBuffer.WARNING, "SEMANTIC ERROR: Husband does not match existing entry for " + gr.family);
                                    }
                                }
                            }
                        } else if(gr.linkto != null) {
                            String other = extractId(gr.linkto);
                            Person spouse = _people.get(other);
                            if("M".equalsIgnoreCase(gp.sex)) {
                                _marriages.put(gr.family, new Marriage(person, spouse, (DateBase)null));
                            } else if("F".equalsIgnoreCase(gp.sex)) {
                                _marriages.put(gr.family, new Marriage(spouse, person, (DateBase)null));
                            }
                        }
                    }
                }
            }
            
            // Set marriage dates
            for(Ged2XmlEvent ge : gp.events) {
                if((ge.type != null) && (ge.date != null)) {
                    if("marriage".equalsIgnoreCase(ge.type) && (ge.family != null)) {
                        Marriage marr = _marriages.get(ge.family);
                        try {
                            marr.setDate(ge.date);
                        } catch(DateException e) {
                            _log.log(LogBuffer.WARNING, e.getMessage());
                        }
                    }
                }
            }
        }
    }


    /**
     * Handles the start of each element.  Creates a new object for those that
     * need one, or sets an property for those that don't.
     */
    public void startElement( String uri, String localName, String qName, Attributes attributes )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case PERSON:
            if( _currPerson != null )
                throw new SAXException( "SEMANTIC ERROR: <PERSON> with Person already open" );
            _currPerson = createPerson( attributes );
            break;
        case SEX:
        case NAME:
        case GIVEN_NAMES:
        case FAMILY_NAME:
        case DATE:
            _tempVal = null;
            break;
        case EVENT:
            if( _currEvent != null )
                throw new SAXException( "SEMANTIC ERROR: <EVENT> with Event already open" );
            _currEvent = createEvent( attributes );
            break;
        case RELATIONSHIP:
            if( _currRelationship != null )
                throw new SAXException( "SEMANTIC ERROR: <RELATIONSHIP> with Relationship already open" );
            _currRelationship = createRelationship( attributes );
            break;
        }
    }
    
    
    /**
     * Handles the end of each element.  Saves any objects that need saving.
     */
    public void endElement( String uri, String localName, String qName )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case PERSON:
            if( _currPerson == null )
                throw new SAXException( "SEMANTIC ERROR: </PERSON> with Person already closed" );
            _currPerson = null;
            break;
        case SEX:
            if( _currPerson == null )
                throw new SAXException( "SEMANTIC ERROR: </SEX> with Person already closed" );
            _currPerson.sex = getText(_tempVal);
            _tempVal = null;
            break;
        case NAME:
            break;
        case GIVEN_NAMES:
            if( _currRelationship == null ) {
                if( _currPerson == null )
                    throw new SAXException( "SEMANTIC ERROR: </GIVEN_NAMES> with Person already closed" );
                if(_currPerson.given_names == null) {
                    _currPerson.given_names = getText(_tempVal);
                } else {
                    _log.log(LogBuffer.WARNING, "WARNING: Person '" + _currPerson.id + "' has additional given name: " + getText(_tempVal) );
                }
            }
            _tempVal = null;
            break;
        case FAMILY_NAME:
            if( _currRelationship == null ) {
                if( _currPerson == null )
                    throw new SAXException( "SEMANTIC ERROR: </FAMILY_NAME> with Person already closed" );
                if(_currPerson.family_name == null) {
                    _currPerson.family_name = getText(_tempVal);
                } else {
                    _log.log(LogBuffer.WARNING, "WARNING: Person '" + _currPerson.id + "' has additional family name: " + getText(_tempVal) );
                }
            }
            _tempVal = null;
            break;
        case EVENT:
            if( _currEvent == null )
                throw new SAXException( "SEMANTIC ERROR: </EVENT> with Event already closed" );
            _currEvent = null;
            break;
        case DATE:
            if( _currEvent != null )
                _currEvent.date = getText(_tempVal);
            _tempVal = null;
            break;
        case RELATIONSHIP:
            if( _currRelationship == null )
                throw new SAXException( "SEMANTIC ERROR: </RELATIONSHIP> with Relationship already closed" );
            _currRelationship = null;
            break;
        }
    }
    
    public void characters(char[] ch, int start, int length)
        throws SAXException
    {
        String text = new String(ch, start, length).trim();
        if(text.length() > 0) {
            if((_tempVal != null) && (_tempVal.length() > 0)) {
                _tempVal += " " + text;
            } else {
                _tempVal = text;
            }
        }
    }

    
    /**
     * Checks that an element is part of the FTML vocabulary and turns it into
     * a token that can be used in a switch statement.  Throws an exception if
     * the element cannot be found.
     */
    private int lookupElement( String qName )
        throws SAXException
    {
        Integer element = (Integer) _elements.get( qName.toLowerCase() );
        
        if( element == null )
            //throw new SAXException( "PARSE ERROR: unrecognised element " + qName );
            return -1;
        
        return element.intValue();
    }
    
    
    private Ged2XPerson createPerson( Attributes attrs )
        throws SAXException
    {
        Ged2XPerson ret = new Ged2XPerson();
        ret.id = attrs.getValue( "idno" );     
        if(ret.id == null)
            throw new SAXException("PARSE ERROR: Person with no idno attribute");
        _gedPeople.put(ret.id, ret);
        return ret;
    }
    
    private Ged2XmlEvent createEvent( Attributes attrs )
        throws SAXException
    {
        if(_currPerson == null)
            throw new SAXException("PARSE ERROR: Event with Person closed");
        Ged2XmlEvent ret = new Ged2XmlEvent();
        ret.type = attrs.getValue( "type" );                
        if(ret.type == null)
            throw new SAXException("PARSE ERROR: Event with no type attribute");
        ret.family = attrs.getValue( "family" );                
        _currPerson.events.add(ret);
        return ret;
    }  
    
    private Ged2XRelationship createRelationship( Attributes attrs )
        throws SAXException
    {
        if(_currPerson == null)
            throw new SAXException("PARSE ERROR: Relationship with Person closed");
        Ged2XRelationship ret = new Ged2XRelationship();
        ret.type = attrs.getValue( "type" );                
        if(ret.type == null)
            throw new SAXException("PARSE ERROR: Relationship with no type attribute");
        ret.linkto = attrs.getValue( "linkto" );                
        ret.family = attrs.getValue( "family" );                
        _currPerson.relationships.add(ret);
        return ret;
    }  
        
    
    private static String getText(String val)
    {
        if(val == null)
            return null;
        StringBuffer ret = new StringBuffer();
        for(int n = 0; n < val.length(); n++) {
            char ch = val.charAt(n);
            if(ch >= 32) {
                ret.append(ch);
            }
        }
        return ret.toString();
    }
    
    /*private static DateBase dateToDateBase( String str )
    {
        String upcase = str.toLowerCase();
        
        if( upcase.startsWith( "before " ) )
            return DateUtil.toDateBase( str.substring( 7 ), ApproxDate.BEFORE, 0 );
        else if( upcase.startsWith( "after " ) )
            return DateUtil.toDateBase( str.substring( 6 ), ApproxDate.AFTER, 0 );
        else if( upcase.startsWith( "about " ) )
            return DateUtil.toDateBase( str.substring( 6 ), ApproxDate.ABOUT, 0 );
        else if( upcase.startsWith( "between " ) )
        {
            String dates = str.substring( str.indexOf(' ') + 1  );
            int and = dates.indexOf( " and " );
            
            if( and != -1 )
            {
                // Split the date on the " and "
                String fromstr = dates.substring( 0, and ).trim();
                String tostr   = dates.substring( and + 5 ).trim();
                
                // Parse the dates
                DateBase from = DateUtil.toDateBase( fromstr );
                DateBase to   = DateUtil.toDateBase( tostr );
                
                if( (from == null) || (to == null) )
                {                    
                    System.out.println( "WARNING: Unable to make sense of date range \""
                                        + dates + "\"" );   
                    return null;
                }
            
                return new DateRange( from, to );
            }
            
            System.out.println( "WARNING: Missing \"and\" in date range \"" + dates + "\"" );
            return null;
        }
        
        return DateUtil.toDateBase( str );
    }*/
    
    private static String extractId(String linkto)
    {
        String id = linkto;
        int dash = linkto.indexOf('-');
        if(dash != -1) id = linkto.substring(0, dash);
        return id;
    }
}