///////////////////////////////////////////////////////////////////////////////
//
// Filename: GedcomXmlHandler.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  27th February 2010
//
// Purpose:  Gedcom XML SAX Handler for use with Xerces.
//
// (c) Copyright R.H.Thornburrow, 2001-10
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.xml.gedcom;

import com.tsac.familytree.*;
import com.tsac.familytree.util.ApproxDate;
import com.tsac.familytree.util.DateBase;
import com.tsac.familytree.util.DateException;
import com.tsac.familytree.util.DateRange;
import com.tsac.familytree.util.DateUtil;
import com.tsac.familytree.util.LogBuffer;
import com.tsac.familytree.xml.IXmlHandlerDelegate;
//import com.tsac.familytree.util.*;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import java.util.Hashtable;
import java.util.Vector;

public class GedcomXmlHandler
    extends org.xml.sax.helpers.DefaultHandler
    implements IXmlHandlerDelegate
{
    private Hashtable<String,Integer> _elements = new Hashtable<String,Integer>();
    private Hashtable<String,Person>  _people    = new Hashtable<String,Person>();
    private Vector<Person>   _roots     = new Vector<Person>();
    private Vector<Marriage> _marriages = new Vector<Marriage>();
    
    private Hashtable<String,GedcomXmlPerson> _gedPeople   = new Hashtable<String,GedcomXmlPerson>();
    private Hashtable<String,GedcomXmlFamily> _gedFamilies = new Hashtable<String,GedcomXmlFamily>();
    private Hashtable<String,GedcomXmlEvent>  _gedEvents   = new Hashtable<String,GedcomXmlEvent>();
    
    private GedcomXmlPerson      _currPerson = null;
    private GedcomXmlFamily      _currFamily = null;
    private GedcomXmlEvent       _currEvent  = null;
    private GedcomXmlParticipant _currParticipant  = null;
    
    private String    _tempVal   = null;
    private Hashtable<String,String> _currLink = new Hashtable<String,String>();
    private LogBuffer _log;
    
    public static final int GEDCOM        = 1;
    public static final int FAMILYREC     = 2;
    public static final int INDIVIDUALREC = 3;
    public static final int EVENTREC      = 4;
    public static final int HUSBFATH      = 21;
    public static final int WIFEMOTH      = 22;
    public static final int CHILD         = 23;
    public static final int INDIVNAME     = 31;
    public static final int GENDER        = 32;
    public static final int PARTICIPANT   = 41;
    public static final int DATE          = 42;
    public static final int LINK          = 100;
    public static final int ROLE          = 411;
    
    
    private class GedcomXmlPerson
    {
        public String id;
        public String name;
        public String gender;
    }
    
    private class GedcomXmlFamily
    {
        public String id;
        public String husb;
        public String wife;
        public Vector<String> children = new Vector<String>();
    }
    
    private class GedcomXmlEvent
    {
        public String id;
        public String type;
        public String date;
        public Vector<GedcomXmlParticipant> participants = new Vector<GedcomXmlParticipant>();
    }

    private class GedcomXmlParticipant
    {
        public Hashtable<String,String> linkattrs = new Hashtable<String,String>();
        public String role;
    }
    
    /**
     * Constructor.
     */
    public GedcomXmlHandler(LogBuffer log)
    {
        _log = log;
        
        _elements.put( "gedcom",        new Integer( GEDCOM ) );
        _elements.put( "familyrec",     new Integer( FAMILYREC ) );
        _elements.put( "individualrec", new Integer( INDIVIDUALREC ) );
        _elements.put( "eventrec",      new Integer( EVENTREC ) );
        _elements.put( "husbfath",      new Integer( HUSBFATH ) );
        _elements.put( "wifemoth",      new Integer( WIFEMOTH ) );
        _elements.put( "child",         new Integer( CHILD ) );
        _elements.put( "indivname",     new Integer( INDIVNAME ) );
        _elements.put( "gender",        new Integer( GENDER ) );
        _elements.put( "participant",   new Integer( PARTICIPANT ) );
        _elements.put( "date",          new Integer( DATE ) );
        _elements.put( "link",          new Integer( LINK ) );
        _elements.put( "role",          new Integer( ROLE ) );
    }


    public Vector<Person> getPeople()  { return new Vector<Person>(_people.values()); }
    public Vector<Person> getRoots()  { return _roots; }
    public Vector<Marriage> getMarriages()  { return _marriages; }


    @Override
    public void endDocument()
        throws SAXException
    {
        // Create all the people objects
        for(GedcomXmlPerson gp : _gedPeople.values()) {
            Person person;

            if( (gp.gender != null) && "M".equalsIgnoreCase( gp.gender ) ) {
                person = new Male( gp.id, "", gp.name );
            } else if( (gp.gender != null) && "F".equalsIgnoreCase( gp.gender ) ) {
                person = new Female( gp.id, "", gp.name );
            } else {
                _log.log(LogBuffer.WARNING, "WARNING: Missing sex for individual " + gp.id);
                person = new Person( gp.id, "", gp.name );
            }
            
            //System.out.println("Person: " + gp.id);
            _people.put( gp.id, person );            
        }
        
        // Now tie up all the family bits and create our Marriage objects
        for(GedcomXmlFamily gf : _gedFamilies.values()) {
            Person husb = null, wife = null;
            if(gf.husb != null)
                husb = _people.get(gf.husb);
            if(gf.wife != null)
                wife = _people.get(gf.wife);
            
            // Now see if we can find a date for the marriage
            String date = null;
            for(GedcomXmlEvent ge : _gedEvents.values()) {
                if( "marriage".equalsIgnoreCase(ge.type) ) {
                    String husbid = null, wifeid = null;
                    for(GedcomXmlParticipant part : ge.participants) {
                        if("husband".equalsIgnoreCase(part.role)) {
                            husbid = part.linkattrs.get("ref");
                        } else if("wife".equalsIgnoreCase(part.role)) {
                            wifeid = part.linkattrs.get("ref");
                        }
                    }
                    if((husbid != null) && (gf.husb != null) && husbid.equals(gf.husb)
                        && (wifeid != null) && (gf.wife != null) && wifeid.equals(gf.wife)) {
                            date = ge.date;
                            break;
                    } else if((husbid != null) && (gf.husb != null) && husbid.equals(gf.husb)
                            && (wifeid == null) && (gf.wife == null)) {
                        date = ge.date;
                        break;
                    } else if((husbid == null) && (gf.husb == null)
                            && (wifeid != null) && (gf.wife != null) && wifeid.equals(gf.wife)) {
                        date = ge.date;
                        break;
                    }
                }
            }
            
            Marriage marr = new Marriage(husb, wife, dateToDateBase(date));
            //System.out.println("Marriages: " + gf.id);
            _marriages.add(marr);
            
            for(String ref : gf.children) {
                Person child = _people.get(ref);
                if(child != null) {
                    if(gf.husb != null)
                        child.setFather(gf.husb);
                    if(gf.wife != null)
                        child.setMother(gf.wife);
                }
            }
        }
        
        // Now apply all the dates from events
        for(GedcomXmlEvent ge : _gedEvents.values()) {
            //System.out.println("Event: " + ge.id + "; " + ge.type + "; " + ge.date);
            if( "marriage".equalsIgnoreCase(ge.type) ) {
                // already processed
            } else {
                for(GedcomXmlParticipant part : ge.participants) {
                    String ref = part.linkattrs.get("ref");
                    if(ref == null) {
                        _log.log(LogBuffer.WARNING, "WARNING: Missing ref on EventRec");
                        continue;
                    }
                    //System.out.println("Ref: " + ref);
                    Person partper = _people.get(ref);
                    if(partper == null) {
                        _log.log(LogBuffer.WARNING, "WARNING: Individual with id '" + ref + "' not found.");
                        continue;
                    }
                    //System.out.println("Participant: " + partper.getName() + "; " + part.role);

                    if( "birth".equalsIgnoreCase(ge.type) ) {
                        if(ge.date != null) {
                            partper.setBorn(dateToDateBase(ge.date));
                        }
                    } else if( "christening".equalsIgnoreCase(ge.type) ) {
                        if("father".equalsIgnoreCase(part.role)) {
                            // Do nothing - should be recorded in family record
                        } else if("mother".equalsIgnoreCase(part.role)) {
                            // Do nothing - should be recorded in family record
                        } else if("child".equalsIgnoreCase(part.role)) {
                            if(ge.date != null) {
                                partper.setChristened(dateToDateBase(ge.date));
                            }
                        }
                    } else if( "death".equalsIgnoreCase(ge.type) ) {
                        if(ge.date != null) {
                            partper.setDied(dateToDateBase(ge.date));
                        }
                    }
                } // for
            }
        }
    }


    /**
     * Handles the start of each element.  Creates a new object for those that
     * need one, or sets an property for those that don't.
     */
    public void startElement( String uri, String localName, String qName, Attributes attributes )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case FAMILYREC:
            if( _currFamily != null )
                throw new SAXException( "SEMANTIC ERROR: <FamilyRec> with FamilyRec already open" );
            _currFamily = createMarriage( attributes );
            break;
        case INDIVIDUALREC:
            if( _currPerson != null )
                throw new SAXException( "SEMANTIC ERROR: <IndividualRec> with IndividualRec already open" );
            _currPerson = createPerson( attributes );
            break;
        case EVENTREC:
            if( _currEvent != null )
                throw new SAXException( "SEMANTIC ERROR: <EventRec> with EventRec already open" );
            _currEvent = createEvent( attributes );
            break;
        case WIFEMOTH:
        case HUSBFATH:
        case CHILD:
            _currLink = null;
            break;
        case LINK:
            _currLink = new Hashtable<String,String>();
            for(int n = 0; n < attributes.getLength(); n++)
                 _currLink.put(attributes.getQName(n).toLowerCase(), attributes.getValue(n));
            break;
        case INDIVNAME:
        case GENDER:
        case ROLE:
        case DATE:
            _tempVal = null;
            break;
        case PARTICIPANT:
            _currParticipant = createParticipant( attributes );
            break;
        }
    }
    
    
    /**
     * Handles the end of each element.  Saves any objects that need saving.
     */
    public void endElement( String uri, String localName, String qName )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case FAMILYREC:
            if( _currFamily == null )
                throw new SAXException( "SEMANTIC ERROR: </FamilyRec> with FamilyRec already closed" );
            _currFamily = null;
            break;
        case INDIVIDUALREC:
            if( _currPerson == null )
                throw new SAXException( "SEMANTIC ERROR: </IndividualRec> with IndividualRec already closed" );
            _currPerson = null;
            break;
        case EVENTREC:
            if( _currEvent == null )
                throw new SAXException( "SEMANTIC ERROR: </EventRec> with EventRec already closed" );
            _currEvent = null;
            break;
        case WIFEMOTH:
            applyMother( _currFamily, _currLink );
            _currLink = null;
            break;
        case HUSBFATH:
            applyFather( _currFamily, _currLink );
            _currLink = null;
            break;
        case CHILD:
            applyChild( _currFamily, _currLink );
            _currLink = null;
            break;
        case INDIVNAME:
            if( _currPerson == null )
                throw new SAXException( "SEMANTIC ERROR: </IndivName> with Person already closed" );
            if(_currPerson.name == null) {
                _currPerson.name = getText(_tempVal);
            } else {
                _log.log(LogBuffer.WARNING, "WARNING: Individual '" + _currPerson.id + "' has additional name: " + getText(_tempVal));
            }
            _tempVal = null;
            break;
        case GENDER:
            if( _currPerson == null )
                throw new SAXException( "SEMANTIC ERROR: </Gender> with Person already closed" );
            _currPerson.gender = getText(_tempVal);
            _tempVal = null;
            break;
        case PARTICIPANT:
            if( _currParticipant == null )
                throw new SAXException( "SEMANTIC ERROR: </Participant> with Participant already closed" );
            _currParticipant.linkattrs = _currLink;
            _currLink = null;
            if( _currEvent == null )
                throw new SAXException( "SEMANTIC ERROR: </Participant> with EventRec already closed" );
            _currEvent.participants.add(_currParticipant);
            _currParticipant = null;
            break;
        case ROLE:
            if( _currParticipant != null )
                _currParticipant.role = getText(_tempVal);
            _tempVal = null;
            break;
        case DATE:
            if( _currEvent != null )
                _currEvent.date = getText(_tempVal);
            _tempVal = null;
            break;
        }
    }
    
    public void characters(char[] ch, int start, int length)
        throws SAXException
    {
        String text = new String(ch, start, length).trim();
        if(text.length() > 0) {
            if((_tempVal != null) && (_tempVal.length() > 0)) {
                _tempVal += " " + text;
            } else {
                _tempVal = text;
            }
        }
    }

    
    /**
     * Checks that an element is part of the FTML vocabulary and turns it into
     * a token that can be used in a switch statement.  Throws an exception if
     * the element cannot be found.
     */
    private int lookupElement( String qName )
        throws SAXException
    {
        Integer element = (Integer) _elements.get( qName.toLowerCase() );
        
        if( element == null )
            //throw new SAXException( "PARSE ERROR: unrecognised element " + qName );
            return -1;
        
        return element.intValue();
    }
    
    
    private GedcomXmlPerson createPerson( Attributes attrs )
        throws SAXException
    {
        GedcomXmlPerson ret = new GedcomXmlPerson();
        ret.id = attrs.getValue( "id" );     
        if(ret.id == null)
            ret.id = attrs.getValue( "Id" );    
        if(ret.id == null)
            throw new SAXException("PARSE ERROR: FamilyRec with no id attribute");
        _gedPeople.put(ret.id, ret);
        return ret;
    }
    
    private GedcomXmlFamily createMarriage( Attributes attrs )
        throws SAXException
    {
        GedcomXmlFamily ret = new GedcomXmlFamily();
        ret.id = attrs.getValue( "id" );    
        if(ret.id == null)
            ret.id = attrs.getValue( "Id" );    
        if(ret.id == null)
            throw new SAXException("PARSE ERROR: FamilyRec with no id attribute");
        _gedFamilies.put(ret.id, ret);
        return ret;
    }
    
    private GedcomXmlEvent createEvent( Attributes attrs )
        throws SAXException
    {
        GedcomXmlEvent ret = new GedcomXmlEvent();
        ret.id = attrs.getValue( "id" );                
        if(ret.id == null)
            ret.id = attrs.getValue( "Id" );    
        if(ret.id == null)
            throw new SAXException("PARSE ERROR: FamilyRec with no id attribute");
        ret.type = attrs.getValue( "type" );                
        if(ret.type == null)
            ret.type = attrs.getValue( "Type" );    
        _gedEvents.put(ret.id, ret);
        return ret;
    }  
    
    private GedcomXmlParticipant createParticipant( Attributes attrs )
        throws SAXException
    {
        GedcomXmlParticipant ret = new GedcomXmlParticipant();
        return ret;
    }  
    
    
    private void applyMother( GedcomXmlFamily family, Hashtable<String,String> linkattrs )
        throws SAXException
    {
        if( family == null )
            throw new SAXException( "SEMANTIC ERROR: <WifeMoth> with FamilyRec already closed" );
            
        family.wife = linkattrs.get( "ref" );
    }
    
    
    private void applyFather( GedcomXmlFamily family, Hashtable<String,String> linkattrs )
        throws SAXException
    {
        if( family == null )
            throw new SAXException( "SEMANTIC ERROR: <HusbFath> with FamilyRec already closed" );
                    
        family.husb = linkattrs.get( "ref" );
    }
    
    
    private void applyChild( GedcomXmlFamily family, Hashtable<String,String> linkattrs )
        throws SAXException
    {
        if( family == null )
            throw new SAXException( "SEMANTIC ERROR: <Child> with FamilyRec already closed" );

        String ref = linkattrs.get( "ref" );
        if(ref == null)
            throw new SAXException("PARSE ERROR: <Link> with no Ref attribute");
        
        family.children.add( ref );
    }
    
    
    private static String getText(String val)
    {
        if(val == null)
            return null;
        StringBuffer ret = new StringBuffer();
        for(int n = 0; n < val.length(); n++) {
            char ch = val.charAt(n);
            if(ch >= 32) {
                ret.append(ch);
            }
        }
        return ret.toString();
    }
    
    private DateBase dateToDateBase( String str )
    {
        if(str == null)
            return null;
        
        String upcase = str.toLowerCase();
        
        try {
            if( upcase.startsWith( "before " ) )
                return DateUtil.toDateBase( str.substring( 7 ), ApproxDate.BEFORE, 0 );
            else if( upcase.startsWith( "after " ) )
                return DateUtil.toDateBase( str.substring( 6 ), ApproxDate.AFTER, 0 );
            else if( upcase.startsWith( "about " ) || upcase.startsWith( "abt " ) )
                return DateUtil.toDateBase( str.substring( 6 ), ApproxDate.ABOUT, 0 );
            else if( upcase.startsWith( "between " ) )
            {
                String dates = str.substring( str.indexOf(' ') + 1  );
                int and = dates.indexOf( " and " );

                if( and != -1 )
                {
                    // Split the date on the " and "
                    String fromstr = dates.substring( 0, and ).trim();
                    String tostr   = dates.substring( and + 5 ).trim();

                    // Parse the dates
                    DateBase from = DateUtil.toDateBase( fromstr );
                    DateBase to   = DateUtil.toDateBase( tostr );

                    if( (from == null) || (to == null) )
                    {                    
                        System.out.println( "WARNING: Unable to make sense of date range \""
                                + dates + "\"" );   
                        return null;
                    }

                    return new DateRange( from, to );
                }

                System.out.println( "WARNING: Missing \"and\" in date range \"" + dates + "\"" );
                return null;
            }

            return DateUtil.toDateBase( str );
        } catch(DateException e) {
            _log.log(LogBuffer.WARNING, e.getMessage());
        }
        
        return null;
    }
}